from pathlib import Path
import tests
from e4s_cl.cf.containers import (
    BackendUnsupported,
    BoundFile,
    Container,
    FileOptions,
    optimize_bind_addition,
)


class ContainerTest(tests.TestCase):

    def test_create(self):
        container = Container(name='dummy')
        self.assertFalse(type(container) == Container)
        self.assertTrue(isinstance(container, Container))

    def test_backend_unknown(self):
        with self.assertRaises(BackendUnsupported):
            container = Container(name='UNKNOWN')

    def test_bind_file(self):
        container = Container(name='dummy')

        target = Path('/tmp')
        dest = Path('/etc')
        contained_dest = Path("/etc/skel")
        option = FileOptions.READ_WRITE

        container.bind_file(target)
        self.assertIn(BoundFile(target, target, FileOptions.READ_ONLY),
                      list(container.bound))

        container._Container__bound_files = {}

        container.bind_file(target, dest=dest)
        self.assertIn(BoundFile(target, dest, FileOptions.READ_ONLY),
                      list(container.bound))

        container._Container__bound_files = {}

        container.bind_file(target, dest=dest)
        self.assertIn(BoundFile(target, dest, FileOptions.READ_ONLY),
                      list(container.bound))

        container.bind_file(target, dest=dest, option=option)
        self.assertIn(BoundFile(target, dest, FileOptions.READ_WRITE),
                      list(container.bound))

        container._Container__bound_files = {}

        container.bind_file(target, dest=contained_dest)
        self.assertIn(BoundFile(target, contained_dest, FileOptions.READ_ONLY),
                      list(container.bound))

        container.bind_file(target, dest=dest, option=option)
        self.assertIn(BoundFile(target, dest, FileOptions.READ_WRITE),
                      list(container.bound))

    def test_bind_file_inclusion(self):
        pmi = BoundFile(Path('/usr/lib/libpmi.so'), Path('/usr/lib/libpmi.so'),
                        FileOptions.READ_ONLY)
        mpi = BoundFile(Path('/usr/lib/libmpi.so'), Path('/usr/lib/libmpi.so'),
                        FileOptions.READ_ONLY)
        mpi_translated = BoundFile(Path('/usr/lib/libmpi.so'),
                                   Path('/.e4s-cl/hostlibs/libmpi.so'),
                                   FileOptions.READ_ONLY)
        usr = BoundFile(Path('/usr'), Path('/usr'), FileOptions.READ_ONLY)

        external_lib = BoundFile(Path('/opt/openmpi/lib'),
                                 Path('/usr/openmpi'), FileOptions.READ_ONLY)

        bound = {mpi, mpi_translated}

        self.assertSetEqual(optimize_bind_addition(pmi, bound), {pmi} | bound)
        self.assertSetEqual(optimize_bind_addition(mpi, bound), bound)
        self.assertSetEqual(optimize_bind_addition(usr, bound),
                            {usr, mpi_translated})

        self.assertSetEqual(optimize_bind_addition(mpi, {usr}), {usr})
        self.assertSetEqual(optimize_bind_addition(mpi_translated, {usr}),
                            {mpi_translated, usr})

        self.assertSetEqual(optimize_bind_addition(external_lib, {usr}),
                            {external_lib, usr})

    def test_bind_relative(self):
        container = Container(name='dummy')

        target = Path('/tmp/../proc/meminfo')

        ref = Path('/tmp')
        file = Path('/proc/meminfo')

        container.bind_file(target)
        files = set(map(lambda x: x.origin, container.bound))

        self.assertSetEqual({ref, file}, files)

    def test_double_bind(self):
        container = Container(name='dummy')

        library = Path(tests.ASSETS, 'libgver.so.0')
        library_symlink = Path(tests.ASSETS, 'libgver.so.0.0.0')
        target = Path('/', 'hostlibs', 'libgver.so.0')

        bind1 = BoundFile(library, target)
        bind2 = BoundFile(library_symlink, target)

        self.assertSetEqual({bind1}, optimize_bind_addition(bind2, {bind1}))
        self.assertSetEqual({bind2}, optimize_bind_addition(bind1, {bind2}))



"""
Base template for commands
"""

from abc import ABCMeta, abstractmethod
from e4s_cl import logger, cli
from e4s_cl.cli.arguments import ArgumentsNamespace


class AbstractCommand(metaclass=ABCMeta):
    """Abstract base class for E4S Container Launcher commands.
    
    Attributes:
        module_name (str): Name of the command module this command object belongs to.
        command (str): Command line string that executes this command.
        summary (str): One-line summary of the command.
        help_page (str): Long and informative description of the command.
        group (str): If not None, commands will be grouped together by group name in help messages.
    """
    def __init__(self,
                 module_name,
                 format_fields=None,
                 summary_fmt=None,
                 help_page_fmt=None,
                 group=None):
        if not summary_fmt:
            summary_fmt = "No summary for '%(command)s'"
        self.module_name = module_name
        self.logger = logger.get_logger(module_name)
        self.command = cli.command_from_module_name(module_name)

        self.format_fields = format_fields or {}

        default_format_fields = [
            ('prog', cli.SCRIPT_COMMAND),
            ('command', self.command),
        ]

        for key, value in default_format_fields:
            self.format_fields.setdefault(key, value)

        self.summary_fmt = summary_fmt
        self.help_page_fmt = help_page_fmt
        self.group = group
        self._parser = None

    def __str__(self):
        return self.command

    @property
    def summary(self):
        return self.summary_fmt % self.format_fields

    @property
    def help_page(self):
        return self.help_page_fmt % self.format_fields if self.help_page_fmt else None

    @property
    def parser(self):
        if self._parser is None:
            self._parser = self._construct_parser()
        return self._parser

    @property
    def usage(self):
        return self.parser.format_help()

    @property
    def monicker(self):
        return self.module_name.split('.')[-1]

    def _parse_args(self, argv):
        if isinstance(argv, ArgumentsNamespace):
            args = argv
        else:
            args = self.parser.parse_args(args=argv)
        self.logger.debug('%s args: %s', self.command, args)
        return args

    @abstractmethod
    def _construct_parser(self):
        """Construct a command line argument parser."""

    @abstractmethod
    def main(self, argv):
        """Command program entry point.
        
        Args:
            argv (list): Command line arguments.
            
        Returns:
            int: Process return code: non-zero if a problem occurred, 0 otherwise
        """



import os
import sys
from pathlib import Path
try:
    from e4s_cl.version import __version__
except ModuleNotFoundError:
    __version__ = "0.0.0"

E4S_CL_VERSION = __version__
"""str: E4S Container Launcher Version"""

EXIT_FAILURE = -100
"""int: Process exit code indicating unrecoverable failure."""

EXIT_WARNING = 100
"""int: Process exit code indicating non-optimal condition on exit."""

EXIT_SUCCESS = 0
"""int: Process exit code indicating successful operation."""

INIT_TEMP_PROFILE_NAME = '__INIT_TEMP_PROFILE'
"""string: Default profile name at initialisation."""

MIN_PYTHON_VERSION = (3, 6)
"""tuple: Required Python version for E4S Container Launcher.

A tuple of at least (MAJOR, MINOR) directly comparible to :any:`sys.version_info`
"""

PYTHON_VERSION = sys.version_info[0:3]
"""list: Version of the running python interpreter"""

if sys.version_info[0] < MIN_PYTHON_VERSION[0] or sys.version_info[
        1] < MIN_PYTHON_VERSION[1]:
    VERSION = '.'.join([str(x) for x in sys.version_info[0:3]])
    EXPECTED = '.'.join([str(x) for x in MIN_PYTHON_VERSION])
    sys.stderr.write(f"""{sys.executable}
{sys.version}
Your Python version is {VERSION} but Python {EXPECTED} is required.
Please install the required Python version or raise an issue on Github for support.
""")
    sys.exit(EXIT_FAILURE)

E4S_CL_TEST = bool(os.environ.get('__E4S_CL_TEST__', False))
"""bool: True if the package is run in a test environment"""

# Use a file to brand installation directories
_HOME_MARKER = '.e4s-cl-home'
__install_home = None

# Look at all the parent of the package to look for the marker
for parent in Path(__file__).resolve().parents:
    if Path(parent, _HOME_MARKER).exists():
        __install_home = parent

E4S_CL_HOME = Path(
    os.environ.get('__E4S_CL_HOME__', __install_home
                   or Path(__file__).parents[1])).resolve().as_posix()
"""str: Absolute path to the top-level E4S Container Launcher directory.

This directory contains at least `bin` and `conda` directories and is the root
for system-level package installation paths. **Do not** change it once it is set.
"""

E4S_CL_ENV_PREFIX = 'E4S_CL'

def _get_e4s_cl_script():
    """Get a reliable path to the e4s-cl script for subprocess invocation."""
    import shutil
    
    # First try the environment variable
    if '__E4S_CL_SCRIPT__' in os.environ:
        script = os.environ['__E4S_CL_SCRIPT__']
        # If it's just a command name, try to resolve it to a full path
        if not os.path.isabs(script) and not os.path.exists(script):
            # Try to find it in PATH
            resolved = shutil.which(script)
            if resolved:
                return resolved
        return script
    
    # Try using sys.argv[0] if it's available
    script_candidate = sys.argv[0] or ''
    if script_candidate:
        return script_candidate
    
    # Fallback to just 'e4s-cl'
    return 'e4s-cl'

E4S_CL_SCRIPT = _get_e4s_cl_script()
"""str: Script that launched E4S Container Launcher.

Mainly used for help messages. **Do not** change it once it is set.
"""

E4S_CL_MPI_TESTER_SCRIPT_NAME = "e4s-cl-mpi-tester"
"""str: Name of the MPI tester script bundled with e4s-cl.
"""

SYSTEM_PREFIX = os.path.realpath(
    os.path.abspath(
        os.environ.get('__E4S_CL_SYSTEM_PREFIX__',
                       os.path.join(E4S_CL_HOME, 'system'))))
"""str: System-level E4S Container Launcher files."""

USER_PREFIX = os.path.abspath(
    os.environ.get(
        '__E4S_CL_USER_PREFIX__',
        os.path.join(os.path.expanduser('~'), '.local', 'e4s_cl')))
"""str: User-level E4S Container Launcher files."""

_CONTAINER_DIR_ENV = os.environ.get("E4S_CL_CONTAINER_DIR") or os.environ.get("E4S_CL_CONTAINER_DIRECTORY")
CONTAINER_DIR = Path(_CONTAINER_DIR_ENV).as_posix() if _CONTAINER_DIR_ENV else Path("/", ".e4s-cl").as_posix()
"""str: Path of a directory in which to bind files when in containers"""

CONTAINER_SCRIPT = Path(CONTAINER_DIR, "script").as_posix()
"""str: Path of the script to execute in the container"""

BAREBONES_LIBRARY_DIR = Path(USER_PREFIX, "barebones_libraries").as_posix()
"""str: Path of the script to execute in the use of the barebones backend"""

BAREBONES_SCRIPT = Path(BAREBONES_LIBRARY_DIR, "barebones_script").as_posix()
"""str: Path of the script to execute in the use of the barebones backend"""

CONTAINER_LIBRARY_DIR = Path(CONTAINER_DIR, "hostlibs").as_posix()
"""str: Path of the libraries bound in the container"""

CONTAINER_BINARY_DIR = Path(CONTAINER_DIR, "executables").as_posix()
"""str: Path of the libraries bound in the container"""

PROFILE_LIST_DEFAULT_COLUMNS = ["selected", "name", "backend", "image"]
"""list[str] columns to display in profile list by default"""

WI4MPI_DIR = Path(USER_PREFIX) / "wi4mpi"
"""Directory in which Wi4MPI releases and build will be put if needed"""

WI4MPI_DEFAULT_INSTALL_DIR = WI4MPI_DIR / 'install'
"""Default installation directory for Wi4MPI"""


def version_banner():
    """Return a human readable text banner describing the E4S Container Launcher installation."""
    import platform
    import socket
    from datetime import datetime
    import e4s_cl.logger
    fmt = ("E4S Container Launcher [ http://e4s-project.github.io/e4s-cl ]\n"
           "\n"
           "Prefix         : %(prefix)s\n"
           "Version        : %(version)s\n"
           "Timestamp      : %(timestamp)s\n"
           "Hostname       : %(hostname)s\n"
           "Platform       : %(platform)s\n"
           "Working Dir.   : %(cwd)s\n"
           "Terminal Size  : %(termsize)s\n"
           "Frozen         : %(frozen)s\n"
           "Python         : %(python)s\n"
           "Python Version : %(pyversion)s\n"
           "Python Impl.   : %(pyimpl)s\n"
           "PYTHONPATH     : %(pythonpath)s\n")
    data = {
        "prefix": E4S_CL_HOME,
        "version": E4S_CL_VERSION,
        "timestamp": str(datetime.now()),
        "hostname": socket.gethostname(),
        "platform": platform.platform(),
        "cwd": os.getcwd(),
        "termsize": 'x'.join([str(dim) for dim in e4s_cl.logger.TERM_SIZE]),
        "frozen": getattr(sys, 'frozen', False),
        "python": sys.executable,
        "pyversion": platform.python_version(),
        "pyimpl": platform.python_implementation(),
        "pythonpath": os.pathsep.join(sys.path)
    }
    return fmt % data



import shlex
import tests
from e4s_cl.model.profile import Profile
from e4s_cl.cli.commands.profile.diff import COMMAND


class ProfileDiffTest(tests.TestCase):
    def tearDown(self):
        Profile.controller().unselect()
        self.resetStorage()

    def test_diff(self):
        Profile.controller().create({'name': 'lhs', 'files': ['/tmp/lhs_only', '/tmp/both']})
        Profile.controller().create({'name': 'rhs', 'files': ['/tmp/rhs_only', '/tmp/both']})

        stdout, _ = self.assertCommandReturnValue(0, COMMAND, shlex.split("lhs rhs"))
        self.assertNotIn('/tmp/both', stdout)
        self.assertIn('< /tmp/lhs_only', stdout)
        self.assertIn('> /tmp/rhs_only', stdout)

    def test_diff_missing_arg(self):
        Profile.controller().create({'name': 'lhs', 'files': ['/tmp/lhs_only', '/tmp/both']})

        self.assertNotCommandReturnValue(0, COMMAND, shlex.split("lhs"))



"""
Tests for e4s_cl/cf/compiler.py - vendor detection from ELF .comment sections.

The per-vendor check functions are pure string predicates, so they are tested
directly.  compiler_vendor() is tested through unittest.mock.patch so that
these tests are independent of any binary on the test host.  _get_comment()
error-handling paths are exercised with synthetic inputs that require no
specific compiler toolchain to be installed.
"""

import os
import tempfile
from pathlib import Path
from unittest.mock import patch

import tests
from e4s_cl.cf.compiler import (
    _gnu_check,
    _llvm_check,
    _intel_check,
    _amd_check,
    _pgi_check,
    _armclang_check,
    _fujitsu_check,
    CompilerVendor,
    compiler_vendor,
    _get_comment,
    available_compilers,
)


class CompilerCheckFunctionTests(tests.TestCase):
    """Unit tests for the pure per-vendor string-predicate functions."""

    def test_gnu_check_true(self):
        self.assertTrue(_gnu_check('GCC: (GNU) 11.3.0 20220519'))

    def test_gnu_check_false(self):
        self.assertFalse(_gnu_check('clang version 14.0.0'))

    def test_llvm_check_true(self):
        self.assertTrue(_llvm_check('clang version 14.0.0 (http://llvm.org/git/clang.git)'))

    def test_llvm_check_false(self):
        self.assertFalse(_llvm_check('GCC: (GNU) 11.3.0'))

    def test_amd_check_true(self):
        self.assertTrue(_amd_check('AMD clang version 15.0.0 (ROCm)'))

    def test_amd_check_false(self):
        self.assertFalse(_amd_check('GCC: (GNU) 11.3.0'))

    def test_intel_check_always_false(self):
        # Current implementation is a stub that always returns False
        self.assertFalse(_intel_check('Intel(R) C++ Compiler 2021.9.0'))
        self.assertFalse(_intel_check(''))

    def test_pgi_check_always_false(self):
        self.assertFalse(_pgi_check('PGI Workstation 21.7'))
        self.assertFalse(_pgi_check(''))

    def test_armclang_check_always_false(self):
        self.assertFalse(_armclang_check('Arm C Compiler armclang 21.1'))
        self.assertFalse(_armclang_check(''))

    def test_fujitsu_check_always_false(self):
        self.assertFalse(_fujitsu_check('Fujitsu Fortran Compiler 4.7.0'))
        self.assertFalse(_fujitsu_check(''))


class CompilerVendorTests(tests.TestCase):
    """Tests for compiler_vendor() using a patched _get_comment.

    No binary files are required; vendor detection logic is completely isolated.
    """

    def test_gnu_binary(self):
        with patch('e4s_cl.cf.compiler._get_comment',
                   return_value='GCC: (GNU) 11.3.0'):
            result = compiler_vendor(Path('/fake/binary'))
        self.assertEqual(result, CompilerVendor.GNU)

    def test_llvm_binary(self):
        with patch('e4s_cl.cf.compiler._get_comment',
                   return_value='clang version 14.0.0'):
            result = compiler_vendor(Path('/fake/binary'))
        self.assertEqual(result, CompilerVendor.LLVM)

    def test_amd_binary(self):
        # AMD comment contains 'AMD', 'clang', and 'GCC' - AMD must win (highest precedence)
        with patch('e4s_cl.cf.compiler._get_comment',
                   return_value='AMD clang version 15.0 (ROCm) - GCC 11.3'):
            result = compiler_vendor(Path('/fake/binary'))
        self.assertEqual(result, CompilerVendor.AMD)

    def test_llvm_over_gnu_precedence(self):
        # clang binaries may embed a GCC version string; LLVM must have priority over GNU
        with patch('e4s_cl.cf.compiler._get_comment',
                   return_value='clang version 14.0 - GCC: (GNU) 11.0'):
            result = compiler_vendor(Path('/fake/binary'))
        self.assertEqual(result, CompilerVendor.LLVM)

    def test_empty_comment_defaults_gnu(self):
        # When neither AMD, LLVM, nor any other vendor matches, fall through to GNU default
        with patch('e4s_cl.cf.compiler._get_comment', return_value=''):
            result = compiler_vendor(Path('/fake/binary'))
        self.assertEqual(result, CompilerVendor.GNU)

    def test_unknown_comment_defaults_gnu(self):
        with patch('e4s_cl.cf.compiler._get_comment',
                   return_value='Some unknown toolchain 99.0'):
            result = compiler_vendor(Path('/fake/binary'))
        self.assertEqual(result, CompilerVendor.GNU)


class GetCommentErrorHandlingTests(tests.TestCase):
    """Tests for _get_comment() when the file path is invalid or not an ELF."""

    def test_nonexistent_file_returns_empty(self):
        result = _get_comment(Path('/nonexistent/no_such_binary_4e2s_cl'))
        self.assertEqual(result, '')

    def test_directory_path_returns_empty(self):
        result = _get_comment(Path('/tmp'))
        self.assertEqual(result, '')

    def test_non_elf_text_file_returns_empty(self):
        with tempfile.NamedTemporaryFile(suffix='.txt', delete=False) as fh:
            fh.write(b'This is plain text, not an ELF binary.\n')
            fname = fh.name
        try:
            result = _get_comment(Path(fname))
            self.assertEqual(result, '')
        finally:
            os.unlink(fname)


class AvailableCompilersTests(tests.TestCase):
    """available_compilers() must return a set without raising regardless of what
    is installed.  We do not assert specific vendors because the test environment
    may have any combination of toolchains installed."""

    def test_returns_set(self):
        result = available_compilers()
        self.assertIsInstance(result, (set, frozenset))

    def test_all_values_are_compiler_vendor(self):
        for vendor in available_compilers():
            self.assertIsInstance(vendor, CompilerVendor)



"""
Tests for the real argument tables of each supported launcher (srun, aprun, jsrun).

These tests verify that realistic command lines split correctly into
(launcher, program) pairs using the actual ARGUMENTS dictionary defined in
each launcher module - not a synthetic test module.  A wrong nargs entry in
a table would silently pass a program argument on the launcher side, so this
catches regressions that are invisible to the generic parser tests.
"""

from shlex import split
import tests
from e4s_cl.cf.launchers import interpret


class SrunParserTest(tests.TestCase):
    """Realistic srun command lines parsed with the real slurm ARGUMENTS table."""

    def test_basic_task_count(self):
        """srun -n 4 ./app"""
        launcher, command = interpret(split('srun -n 4 ./app'))
        self.assertEqual(launcher, ['srun', '-n', '4'])
        self.assertEqual(command, ['./app'])

    def test_ntasks_long_equals(self):
        """srun --ntasks=4 ./app"""
        launcher, command = interpret(split('srun --ntasks=4 ./app'))
        self.assertIn('--ntasks=4', launcher)
        self.assertEqual(command, ['./app'])

    def test_ntasks_long_space(self):
        """srun --ntasks 4 ./app"""
        launcher, command = interpret(split('srun --ntasks 4 ./app'))
        self.assertIn('--ntasks', launcher)
        self.assertIn('4', launcher)
        self.assertEqual(command, ['./app'])

    def test_nodes_and_tasks(self):
        """srun -N 2 -n 4 ./app"""
        launcher, command = interpret(split('srun -N 2 -n 4 ./app'))
        self.assertEqual(launcher, ['srun', '-N', '2', '-n', '4'])
        self.assertEqual(command, ['./app'])

    def test_account_flag(self):
        """srun -A myaccount ./app"""
        launcher, command = interpret(split('srun -A myaccount ./app'))
        self.assertEqual(launcher, ['srun', '-A', 'myaccount'])
        self.assertEqual(command, ['./app'])

    def test_partition_flag(self):
        """srun -p gpu ./app"""
        launcher, command = interpret(split('srun -p gpu ./app'))
        self.assertIn('gpu', launcher)
        self.assertEqual(command, ['./app'])

    def test_double_dash_separator(self):
        """srun -n 4 -- ./app --bar  (program args don't bleed into launcher)

        interpret() strips '--' from both sides; everything before it is the
        launcher and everything after is the program.
        """
        launcher, command = interpret(
            ['srun', '-n', '4', '--', './app', '--bar'])
        self.assertIn('srun', launcher)
        self.assertNotIn('./app', launcher)
        self.assertEqual(command, ['./app', '--bar'])

    def test_concatenated_n(self):
        """srun -n4 ./app  (concatenated known short-arg)"""
        launcher, command = interpret(split('srun -n4 ./app'))
        self.assertIn('srun', launcher)
        # The program must not be swallowed into the launcher
        self.assertEqual(command, ['./app'])

    def test_gpu_flags(self):
        """srun --gpus-per-task=1 -n 4 ./app"""
        launcher, command = interpret(split('srun --gpus-per-task=1 -n 4 ./app'))
        self.assertIn('--gpus-per-task=1', launcher)
        self.assertEqual(command, ['./app'])

    def test_job_name(self):
        """srun --job-name=myjob -n 2 ./app"""
        launcher, command = interpret(split('srun --job-name=myjob -n 2 ./app'))
        self.assertIn('--job-name=myjob', launcher)
        self.assertEqual(command, ['./app'])

    def test_program_with_args(self):
        """srun -n 2 ./app arg1 arg2"""
        launcher, command = interpret(split('srun -n 2 ./app arg1 arg2'))
        self.assertEqual(launcher, ['srun', '-n', '2'])
        self.assertEqual(command, ['./app', 'arg1', 'arg2'])


class AprunParserTest(tests.TestCase):
    """Realistic aprun (Cray ALPS) command lines."""

    def test_basic_width(self):
        """aprun -n 4 ./app"""
        launcher, command = interpret(split('aprun -n 4 ./app'))
        self.assertEqual(launcher, ['aprun', '-n', '4'])
        self.assertEqual(command, ['./app'])

    def test_pes_per_node(self):
        """aprun -n 4 -N 2 ./app"""
        launcher, command = interpret(split('aprun -n 4 -N 2 ./app'))
        self.assertEqual(launcher, ['aprun', '-n', '4', '-N', '2'])
        self.assertEqual(command, ['./app'])

    def test_depth_flag(self):
        """aprun -n 4 -d 4 ./app"""
        launcher, command = interpret(split('aprun -n 4 -d 4 ./app'))
        self.assertEqual(launcher, ['aprun', '-n', '4', '-d', '4'])
        self.assertEqual(command, ['./app'])

    def test_double_dash(self):
        """aprun double-dash separator - interpret() strips '--' from both sides."""
        launcher, command = interpret(['aprun', '-n', '4', '--', './app'])
        self.assertIn('aprun', launcher)
        self.assertNotIn('./app', launcher)
        self.assertEqual(command, ['./app'])

    def test_memory_per_pe(self):
        """aprun -n 4 -m 512m ./app"""
        launcher, command = interpret(split('aprun -n 4 -m 512m ./app'))
        self.assertIn('-m', launcher)
        self.assertEqual(command, ['./app'])

    def test_program_with_args(self):
        """aprun -n 4 ./app --config foo"""
        launcher, command = interpret(split('aprun -n 4 ./app --config foo'))
        self.assertEqual(launcher, ['aprun', '-n', '4'])
        self.assertEqual(command, ['./app', '--config', 'foo'])


class JsrunParserTest(tests.TestCase):
    """Realistic jsrun (IBM LSF) command lines."""

    def test_basic_nrs(self):
        """jsrun -n 4 ./app"""
        launcher, command = interpret(split('jsrun -n 4 ./app'))
        self.assertEqual(launcher, ['jsrun', '-n', '4'])
        self.assertEqual(command, ['./app'])

    def test_resource_set(self):
        """jsrun --nrs 4 --cpu_per_rs 2 ./app"""
        launcher, command = interpret(
            split('jsrun --nrs 4 --cpu_per_rs 2 ./app'))
        self.assertIn('--nrs', launcher)
        self.assertIn('4', launcher)
        self.assertIn('--cpu_per_rs', launcher)
        self.assertEqual(command, ['./app'])

    def test_gpu_per_rs(self):
        """jsrun --nrs 2 --gpu_per_rs 1 ./app"""
        launcher, command = interpret(split('jsrun --nrs 2 --gpu_per_rs 1 ./app'))
        self.assertIn('--gpu_per_rs', launcher)
        self.assertEqual(command, ['./app'])

    def test_double_dash(self):
        """jsrun double-dash separator - interpret() strips '--' from both sides."""
        launcher, command = interpret(['jsrun', '-n', '4', '--', './app'])
        self.assertIn('jsrun', launcher)
        self.assertNotIn('./app', launcher)
        self.assertEqual(command, ['./app'])

    def test_program_with_args(self):
        """jsrun -n 2 ./app -x 1"""
        launcher, command = interpret(split('jsrun -n 2 ./app -x 1'))
        self.assertEqual(launcher, ['jsrun', '-n', '2'])
        self.assertEqual(command, ['./app', '-x', '1'])



"""
Profile data model.
"""

from pathlib import Path
from e4s_cl import logger
from e4s_cl.error import ProfileSelectionError
from e4s_cl.mvc.model import Model
from e4s_cl.mvc.controller import Controller
from e4s_cl.cf.storage.levels import USER_STORAGE

LOGGER = logger.get_logger(__name__)


def attributes():
    return {
        'name': {
            'primary_key': True,
            'type': 'string',
            'unique': True,
            'description': 'profile name',
        },
        "wi4mpi": {
            'type': 'string',
            'description': 'root of the wi4mpi installation to use (optional)',
        },
        'backend': {
            'type': 'string',
            'description': 'container backend technology',
        },
        'image': {
            'type': 'string',
            'description': 'image identifier',
        },
        'files': {
            'type': 'list',
            'description': 'files to bind in the container',
        },
        'libraries': {
            'type': 'list',
            'description': 'libraries to bind in the container',
        },
        'source': {
            'type': 'string',
            'description': 'script to source before execution (optional)',
        },
    }


def homogenize_files(data: dict) -> None:
    """Cast all paths to posix-compliant format"""
    if not isinstance(data, dict) or 'files' not in data:
        return

    def _sanitize(path: str) -> str:
        return Path(path).as_posix()

    data['files'] = list(map(_sanitize, data.get('files', [])))


class ProfileController(Controller):
    """Profile data controller."""

    def create(self, data):
        homogenize_files(data)
        return super().create(data)

    def delete(self, keys):
        to_delete = self.one(keys)

        try:
            selected = self.selected()
        except ProfileSelectionError:
            pass
        else:
            if selected == to_delete:
                self.unselect()

        super().delete(keys)

    def select(self, keys):
        """Mark a profile as selected"""
        to_select = self.one(keys)

        if to_select is not None:
            self.storage['selected_profile'] = to_select.eid
        else:
            raise ProfileSelectionError('No matching profile')

    def unselect(self):
        if self.storage.contains({'key': 'selected_profile'}):
            del self.storage['selected_profile']

    def selected(self):
        try:
            selected = self.one(self.storage['selected_profile'])
            if not selected:
                raise KeyError
        except KeyError as key_err:
            raise ProfileSelectionError("No profile selected") from key_err
        else:
            return selected

    def update(self, data, keys):
        homogenize_files(data)
        super().update(data, keys)


class Profile(Model):
    """Profile data controller."""

    __attributes__ = attributes
    __controller__ = ProfileController

    @classmethod
    def controller(cls, storage=USER_STORAGE):
        return cls.__controller__(cls, storage)

    @classmethod
    def selected(cls, storage=USER_STORAGE):
        try:
            return cls.__controller__(cls, storage).selected()
        except ProfileSelectionError:
            return {}



from os import getcwd, environ, pathsep
from unittest import skipIf
from pathlib import Path
import tests
from e4s_cl import config, BAREBONES_LIBRARY_DIR
from e4s_cl.util import which, list_directory_files
from e4s_cl.cf.containers import (
    BackendUnsupported,
    BoundFile,
    Container,
    FileOptions,
)

CONFIG_EXECUTABLE = tests.ASSETS / 'bin' / 'barebones-conf'
DEFAULT_CONFIGURATION = config.CONFIGURATION
TEST_CONFIGURATION = config.Configuration.create_from_string(f"""
backends:
  barebones:
    options: ['--nocolor', '-s']
    exec_options: ['--hostname', 'XxmycoolcontainerxX']
""")


class ContainerTestBarebones(tests.TestCase):

    def test_create(self):
        container = Container(name='barebones', image='None')
        self.assertFalse(type(container) == Container)
        self.assertTrue(isinstance(container, Container))

    def test_run_mpirun(self):
        container = Container(name='barebones', image='dummyimagename')
        command = ['mpirun -n 2 ls']
        container_cmd = container._prepare(command)
        self.assertIn(command[0], ' '.join(map(str, container_cmd)))

    def test_bind_file(self):
        container = Container(name='barebones')

        target = Path('/tmp')
        dest = Path('/etc')
        contained_dest = Path("/etc/skel")
        option = FileOptions.READ_WRITE

        container.bind_file(target)
        self.assertIn('tmp',
                      [Path(path).name for path in list_directory_files(Path(BAREBONES_LIBRARY_DIR))])

        target = Path('/tmp2')
        dest = Path(BAREBONES_LIBRARY_DIR + str(target))
        container.bind_file(target, dest=dest)
        self.assertIn('tmp2',
                      [Path(path).name for path in list_directory_files(Path(BAREBONES_LIBRARY_DIR))])

    def test_additional_options_config(self):
        container = Container(name='barebones')
        command = ['']

        barebones_command = container._prepare(command)
        for option in {'--nocolor', '-s', '--hostname', 'XxmycoolcontainerxX'}:
            self.assertNotIn(option, barebones_command)

        config.update_configuration(TEST_CONFIGURATION)
        barebones_command = container._prepare(command)
        self.assertContainsInOrder([
            '--nocolor',
            '-s',
            '--hostname',
            'XxmycoolcontainerxX',
        ], barebones_command)

        config.update_configuration(DEFAULT_CONFIGURATION)
        barebones_command = container._prepare(command)
        for option in {'--nocolor', '-s', '--hostname', 'XxmycoolcontainerxX'}:
            self.assertNotIn(option, barebones_command)

    def test_additional_options_environment(self):
        container = Container(name='barebones')
        command = ['']

        barebones_command = container._prepare(command)
        for option in {'--nocolor', '-s', '--hostname', 'XxmycoolcontainerxX'}:
            self.assertNotIn(option, barebones_command)

        environ['E4S_CL_BAREBONES_OPTIONS'] = "--nocolor -s"
        environ[
            'E4S_CL_BAREBONES_EXEC_OPTIONS'] = "--hostname XxmycoolcontainerxX"
        barebones_command = container._prepare(command)
        self.assertContainsInOrder([
            '--nocolor',
            '-s',
            '--hostname',
            'XxmycoolcontainerxX',
        ], barebones_command)

        del environ['E4S_CL_BAREBONES_OPTIONS']
        del environ['E4S_CL_BAREBONES_EXEC_OPTIONS']
        barebones_command = container._prepare(command)
        for option in {'--nocolor', '-s', '--hostname', 'XxmycoolcontainerxX'}:
            self.assertNotIn(option, barebones_command)



"""
Ensure the profile list command functions as intended
"""

import tests
from e4s_cl import config
from e4s_cl.model.profile import Profile
from e4s_cl.cli.commands.profile.list import COMMAND


class ProfileListTest(tests.TestCase):
    """
    Partial class, as methods are added manually below
    """

    def tearDown(self):
        self.resetStorage()

    def test_list(self):
        Profile.controller().create({"name": 'test01'})
        self.assertCommandReturnValue(0, COMMAND, "test01")

    def test_existence(self):
        _, stderr = self.assertNotCommandReturnValue(0, COMMAND, ['test01'])
        self.assertIn('profile list [profile_name] [profile_name]', stderr)
        self.assertIn('profile list: error:', stderr)

    def test_pattern(self):
        Profile.controller().create({"name": 'test01'})
        Profile.controller().create({"name": 'test02'})
        Profile.controller().create({"name": 'otherName01'})
        stdout, _ = self.assertCommandReturnValue(0, COMMAND, ['test0'])
        self.assertIn('test01', stdout)
        self.assertIn('test02', stdout)
        self.assertNotIn('otherName01', stdout)


def wrapper(key, value, test_name):
    """
    Generate tests from a simple pattern to ensure all fields are correctly shown
    """

    def generated_test(self):
        Profile.controller().create({'name': 'test01', key: value})
        stdout, _ = self.assertCommandReturnValue(0, COMMAND, '')
        if isinstance(value, list):
            self.assertIn(str(len(value)), stdout)
        else:
            self.assertIn(value, stdout)

    generated_test.__name__ = f"test_{test_name}"

    return generated_test


_fields = [('name', 'test_name', 'profile_list_name'),
           ('backend', 'test_back', 'profile_list_backend'),
           ('image', 'test_image', 'profile_list_image'),
           ('libraries', ['test_libraries01',
                          'test_libraries02'], 'profile_list_libraries_count'),
           ('files', ['test_files01', 'test_files02',
                      'test_files03'], 'profile_list_files_count')]

for arguments in _fields:
    test = wrapper(*arguments)
    setattr(ProfileListTest, test.__name__, test)



"""
Podman container manager support
"""

import os
from pathlib import Path
from typing import List
from e4s_cl.error import InternalError
from e4s_cl.util import run_subprocess
from e4s_cl.logger import get_logger
from e4s_cl.cf.containers import Container, FileOptions, BackendNotAvailableError

LOGGER = get_logger(__name__)

NAME = 'podman'
MIMES = []


def opened_fds():
    """
    -> set[int]

    Returns a list of all the opened file descriptors opened by the current
    process
    """
    fds = set()

    for file in Path('/proc/self/fd').glob('*'):
        if not file.exists():
            continue

        try:
            fd_no = int(file.name)
        except ValueError:
            continue

        fds.add(fd_no)

    return fds


class FDFiller:
    """
    Context manager that will "fill" the opened file descriptors to have a
    contiguous list, and make every fd inheritable
    """

    def __init__(self):
        """
        Initialize by creating a buffer of opened files
        """
        self.__opened_files = []

    def __enter__(self):
        """
        Create as many open files as necessary
        """
        fds = opened_fds()

        # Make every existing file descriptor inheritable
        for fd in fds:
            try:
                os.set_inheritable(fd, True)
            except OSError as err:
                if err.errno == 9:
                    continue

        # Compute all the missing numbers in the list
        missing = set(range(max(fds))) - set(fds)

        while missing:
            # Open files towards /dev/null
            null = open('/dev/null', 'w', encoding='utf-8')

            if null.fileno() not in missing:
                raise InternalError(f"Unexpected fileno: {null.fileno()}")

            try:
                # Set the file as inheritable
                os.set_inheritable(null.fileno(), True)
            except OSError as err:
                if err.errno == 9:
                    continue

            # It is not missing anymore
            missing.discard(null.fileno())
            self.__opened_files.append(null)

        passed_fds = fds - {0, 1, 2}
        LOGGER.debug("Passing %d file descriptors: (%s)", len(passed_fds), passed_fds)
        LOGGER.debug("Created %d file descriptors: %s",
                     len(self.__opened_files),
                     [f.fileno() for f in self.__opened_files])

    def __exit__(self, type_, value, traceback):
        for file in self.__opened_files:
            file.close()


class PodmanContainer(Container):
    """
    Podman container object
    """

    executable_name = 'podman'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _fd_number(self):
        """
        -> int
        Podman requires the --preserve-fds=K option to pass file descriptors;
        K being the amount (in addition of 0,1,2) of fds to pass. It also is
        strict on the existence and inheritance flag of those descriptors, and
        will not function if any one of them is invalid/uninheritable.
        """

        fds = opened_fds() - {0, 1, 2}

        return len(fds)

    def _format_bound(self):

        def _format():
            for file in self.bound:
                params = {
                    'type': 'bind',
                    'src': file.origin.as_posix(),
                    'dst': file.destination.as_posix(),
                }

                if file.option == FileOptions.READ_ONLY:
                    params['ro'] = 'true'

                yield "--mount=" + ",".join(f"{k}={v}"
                                            for (k, v) in params.items())

        return list(_format())

    def _prepare(self, command: List[str], overload: bool = True) -> List[str]:
        """
        Prepare a command line to run the given command in a podman container
        """

        def _working_dir() -> List[str]:
            cwd = os.getcwd()
            if cwd in map(lambda x: x.origin, self.bound):
                return ['--workdir', cwd]
            return []

        if overload:
            podman_command = [
                *self._additional_options(),  # Additional options
                'run',  # Run a container
                '--rm',  # Remove when done
                '--ipc=host',  # Use host IPC /!\
                '--env-host',  # Pass host environment /!\
                f"--preserve-fds={self._fd_number()}",  # Inherit file descriptors /!\
                *_working_dir(),  # Work in the same CWD
                *self._format_bound(),  # Bound files options
                *self._additional_options('run'),  # Additional run options
                self.image,
                *command,
            ]
        else:
            podman_command = [
                'run',  # Run a container
                '--rm',  # Remove when done
                f"--preserve-fds={self._fd_number()}",  # Inherit file descriptors /!\
                self.image,
                *command,
            ]

        return podman_command

    def run(self, command: List[str], overload: bool = True) -> int:
        executable = self._executable()
        if executable is None:
            raise BackendNotAvailableError(self.__class__.__name__)

        with FDFiller():
            container_cmd = [executable, *self._prepare(command, overload)]
            return run_subprocess(container_cmd, env=self.env)


CLASS = PodmanContainer



"""Unit test initializations and utility functions."""

import os
import sys
import glob
import shutil
import shlex
import atexit
import tempfile
import unittest
from unittest import skipIf, skipUnless
from pathlib import Path
from e4s_cl.util import get_command_output
import warnings
from io import StringIO
from e4s_cl import logger, E4S_CL_HOME, EXIT_SUCCESS, EXIT_FAILURE, config

ASSETS = Path(__file__).parent / "assets"
CONFIGURATION_FILE = ASSETS / "e4s-cl.yaml"

TEST_SYSTEM = '__test_system'

# Set the configuration to a specific testing version
config.update_configuration(
    config.Configuration.create_from_file(CONFIGURATION_FILE))

from e4s_cl.error import ConfigurationError
from e4s_cl.cf.storage.levels import USER_STORAGE, SYSTEM_STORAGE

_NOT_IMPLEMENTED = []


def not_implemented(cls):
    """Decorator for TestCase classes to indicate that the tests have not been written (yet)."""
    msg = "%s: tests have not been implemented" % cls.__name__
    _NOT_IMPLEMENTED.append(msg)
    return unittest.skip(msg)(cls)


def _null_decorator(_):
    return _


class TestCase(unittest.TestCase):
    """Base class for unit tests.

    Performs tests in a temporary directory and reconfigures :any:`e4s_cl.logger` to work with :any:`unittest`.
    """
    # Follow the :any:`unittest` code style.
    # pylint: disable=invalid-name

    @classmethod
    def setUpClass(cls):
        # Reset stdout logger handler to use buffered unittest stdout
        # pylint: disable=protected-access
        cls._orig_stream = logger._STDERR_HANDLER.stream
        logger._STDERR_HANDLER.stream = sys.stdout

        # Make sure the storage is clean before any test is performed
        cls.resetStorage()

    @classmethod
    def tearDownClass(cls):
        # Reset stdout logger handler to use original stdout
        # pylint: disable=protected-access
        logger._STDERR_HANDLER.stream = cls._orig_stream

    def run(self, result=None):
        # Whenever running a test, set the terminal size large enough to avoid any regex failures due to line wrap
        logger.TERM_SIZE = (150, 150)
        logger.LINE_WIDTH = logger.TERM_SIZE[0]
        logger._STDERR_HANDLER.setFormatter(
            logger.LogFormatter(line_width=logger.LINE_WIDTH,
                                printable_only=True))
        # Nasty hack to give us access to what sys.stderr becomes when unittest.TestRunner.buffered == True
        # pylint: disable=attribute-defined-outside-init
        assert result is not None
        self._result_stream = result.stream
        return super(TestCase, self).run(result)

    def exec_command(self, cmd, argv):
        """Execute a e4s_cl command's main() routine and return the exit code, stdout, and stderr data.

        Args:
            cmd (type): A command instance that has a `main` callable attribute.
            argv (list): Arguments to pass to cmd.main()

        Returns:
            tuple: (retval, stdout, stderr) results of running the command.
        """

        if isinstance(argv, str):
            argv = shlex.split(argv)

        # pylint: disable=protected-access
        stdout = tempfile.TemporaryFile(mode='w+', buffering=1)
        stderr = tempfile.TemporaryFile(mode='w+', buffering=1)
        orig_stdout, orig_stderr = sys.stdout, sys.stderr
        try:
            sys.stdout, sys.stderr = stdout, stderr
            logger._STDERR_HANDLER.stream = stderr
            try:
                retval = cmd.main(argv)
            except SystemExit as err:
                retval = err.code
            stdout.seek(0), stderr.seek(0)
            stdout_value, stderr_value = stdout.read(), stderr.read()
            return retval, stdout_value, stderr_value
        finally:
            sys.stdout, sys.stderr = orig_stdout, orig_stderr
            stdout.close(), stderr.close()
            logger._STDERR_HANDLER.stream = orig_stdout

    def assertCommandReturnValue(self, return_value, cmd, argv):
        retval, stdout, stderr = self.exec_command(cmd, argv)
        if retval != return_value:
            print(stderr, file=sys.stderr)
        self.assertEqual(retval, return_value)
        return stdout, stderr

    def assertNotCommandReturnValue(self, return_value, cmd, argv):
        retval, stdout, stderr = self.exec_command(cmd, argv)
        if retval == return_value:
            print(stderr, file=sys.stderr)
        self.assertNotEqual(retval, return_value)
        return stdout, stderr

    def assertContainsInOrder(self, subset, ordered_iterable):
        """Asserts the ordered_iterable contains the elements of the subset,
        in order, with possibly elements in between"""
        matches = list(filter(lambda x: x in subset, ordered_iterable))

        self.assertListEqual(subset, matches)

    @classmethod
    def resetStorage(cls):
        tables = ["Profile"]
        for table in tables:
            USER_STORAGE.purge(table_name=table)
            SYSTEM_STORAGE.purge(table_name=table)


class TestRunner(unittest.TextTestRunner):
    """Test suite runner."""

    def __init__(self, *args, **kwargs):
        super(TestRunner, self).__init__(*args, **kwargs)
        self.buffer = True

    def run(self, test):
        result = super(TestRunner, self).run(test)
        for item in _NOT_IMPLEMENTED:
            print("WARNING: %s" % item)
        if result.wasSuccessful():
            return EXIT_SUCCESS
        return EXIT_FAILURE



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..'))

# -- Project information -----------------------------------------------------

project = 'E4S Container Launcher'
copyright = '2024, Frederick Deny'
author = 'Frederick Deny'

# The full version, including alpha/beta/rc tags
version = release = '1.0.8'

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosectionlabel',
    'sphinx_toolbox.collapse',
    'sphinxcontrib.sass',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = []

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'furo'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]
html_css_files = ["css/main.css"]

sass_src_dir = "styles"
sass_out_dir = "_static/css"
sass_targets = {"main.scss": "main.css"}

# -- HTML theme options ------------------------------------------------------

# Make fonts bigger for clarity
options = {
    "code-font-size": "1em",
    "admonition-title-font-size": "1em",
    "admonition-font-size": ".92em",
    "font-size--small--2": ".90em",
    "font-size--small--3": ".82em",
}

html_theme_options = {
    "navigation_with_keys": True,
    "light_css_variables": options,
    "dark_css_variables": options
}

# -- Options for man output --------------------------------------------------

man_pages = [('index', 'e4s-cl', 'E4S Container Launcher', '', 1)]

# -- Options for code color  -------------------------------------------------

pygments_style = "gruvbox-light"
pygments_dark_style = "gruvbox-dark"



"""ALPS launcher support, from aprun 6.6.59"""

from e4s_cl.cf.launchers import Parser

SCRIPT_NAMES = ['aprun']

ARGUMENTS = {
    ":": 0,
    "-a": 1,  #arch
    "--access-mode": 1,  #flag
    "--architecture": 1,  #arch
    "-b": 0,
    "-B": 0,
    "--batch-args": 0,
    "--bypass-app-transfer": 0,
    "-C": 0,
    "--cc": 1,  #cpu_list
    "--cp": 1,  #file
    "--cpu-binding": 1,  #cpu_list
    "--cpu-binding-file": 1,  #file
    "--cpus-per-cu": 1,  #CPUs
    "--cpus-per-pe": 1,  #depth
    "--cpu-time-limit": 1,  #sec
    "-D": 0,
    "-d": 1,  #depth
    "--debug level": 0,
    "-e": 1,  #env
    "-E": 1,  #node_list
    "--environment-override": 1,  #env
    "--exclude-node-list": 1,  #node_list
    "--exclude-node-list-file": 1,  #node_list_file
    "-F": 1,  #flag
    "--help": 0,
    "-j": 1,  #CPUs
    "-L": 1,  #node_list
    "-l": 1,  #node_list_file
    "-m": 1,  #size
    "--memory-per-pe": 1,  #size
    "--mpmd-env": 1,  #env
    "-N": 1,  #pes
    "-n": 1,  #width
    "--node-list": 1,  #node_list
    "--node-list-file": 1,  #node_list_file
    "-P": 0,
    "-p": 1,  #pdi
    "--pes": 1,  #width
    "--pes-per-node": 1,  #pes
    "--pes-per-numa-node": 1,  #pes
    "--p-governor": 1,  #governor_name
    "--pipes pipes": 0,
    "--protection-domain": 1,  #pdi
    "--p-state": 1,  #pstate
    "-q": 0,
    "--quiet": 0,
    "-r": 1,  #CPUs
    "-R": 1,  #max_shrink
    "--reconnect": 0,
    "--relaunch": 1,  #max_shrink
    "-S": 1,  #pes
    "--specialized-cpus": 1,  #CPUs
    "--ss": 0,
    "--strict-memory-containment": 0,
    "--sync-output": 0,
    "-T": 0,
    "-t": 1,  #sec
    "--version": 0,
    "--wdir": 1,  #wdir
    "-z": 0,
    "-Z": 1,  #secs
    "--zone-sort": 0,
    "--zone-sort-secs": 1,  #secs
}

PARSER = Parser(ARGUMENTS)



"""
Module defining state-altering methods
"""
import os
from argparse import Action

# Status to decide if you have to handle the error messages
CHILD_MARKER = "__E4S_CL_WORKER"

# Dry run mode, print subprocesses instead of running them
DRY_RUN = False


def is_parent():
    return not os.environ.get(CHILD_MARKER, False)


def set_parent():
    os.environ[CHILD_MARKER] = str(1)


class ParentStatus:

    def __enter__(self):
        set_parent()

    def __exit__(self, type_, value, traceback):
        if os.environ.get(CHILD_MARKER):
            os.environ.pop(CHILD_MARKER)


def is_dry_run():
    return DRY_RUN


def set_dry_run(value):
    global DRY_RUN
    DRY_RUN = value


class DryRunAction(Action):

    def __call__(self, parser, namespace, values, option_string=None):
        set_dry_run(True)



#!/usr/bin/env python3
"""Shell completion support for e4s-cl.

Generates a static bash ``complete -F`` completion script from the live
e4s_cl CLI parsers.  The generated script is committed to the repository as
``completions/e4s-cl.bash`` and its freshness is verified in CI.

Runtime design
--------------
* All subcommand names and option flags are **inlined as string literals** in
  the generated bash function → zero subprocess cost for the common case.
* Dynamic profile names are fetched via an embedded ``python3`` heredoc in
  the bash function → subprocess only for profile-name argument slots.

Entry point (``main()``)
------------------------
* No arguments → print the ``source /path`` setup snippet.
* ``--regenerate`` → rebuild ``completions/e4s-cl.bash`` from live parsers.

For users
---------
* ``make install``: Makefile copies ``completions/e4s-cl.bash`` to
  ``COMPLETION_DIR`` (auto-sourced by bash-completion on new shells).
* Git clone / editable install: ``source /repo/completions/e4s-cl.bash``.
"""

from __future__ import annotations

import json
import os
import sys
import textwrap
from pathlib import Path

# ── Profile store path ────────────────────────────────────────────────────────
# Honour the same env-var override used by e4s_cl storage so tests that set
# __E4S_CL_USER_PREFIX__ to a temp dir work correctly.
_DB_PATH = (
    Path(os.environ.get('__E4S_CL_USER_PREFIX__',
                        Path.home() / '.local' / 'e4s_cl'))
    / 'user.json'
)

# Location of the canonical generated bash completion file.
# Path: e4s_cl/scripts/completion.py → ../../completions/e4s-cl.bash
_BASH_FILE = Path(__file__).parent.parent.parent / 'completions' / 'e4s-cl.bash'

# ── ArgumentCount / ExpectedType constants ────────────────────────────────────
_AC_ATMOSTONE  = 'ARGS_ATMOSTONE'
_AC_ATLEASTONE = 'ARGS_ATLEASTONE'
_AC_SOME       = 'ARGS_SOME'
_ET_PROFILE    = 'defined_profile'

_PROFILE_FN = '__e4s_cl_complete_profiles'

# ── Profile loader — no e4s_cl import ────────────────────────────────────────

def _load_profile_names() -> list[str]:
    """Read profile names directly from the tinydb JSON (no e4s_cl import)."""
    try:
        with open(_DB_PATH) as fh:
            data = json.load(fh)
        return [
            v['name']
            for v in data.get('Profile', {}).values()
            if isinstance(v, dict) and 'name' in v
        ]
    except Exception:
        return []


# ── Bash generation helpers ───────────────────────────────────────────────────

def _nargs_takes_arg(nargs) -> bool:
    """Return True if this nargs value means the option consumes ≥1 argument."""
    if nargs is None or nargs == 0:
        return False
    if isinstance(nargs, int) and nargs > 0:
        return True
    return nargs in (_AC_ATMOSTONE, _AC_ATLEASTONE, _AC_SOME)


def _all_option_names(command: dict) -> list[str]:
    return [name for opt in command.get('options', [])
            for name in opt.get('names', [])]


def _profile_option_patterns(command: dict) -> list[str]:
    """Return option flags whose argument expects a profile name."""
    names: list[str] = []
    for opt in command.get('options', []):
        if (opt.get('expected_type') == _ET_PROFILE
                and _nargs_takes_arg(opt.get('arguments'))):
            names.extend(opt.get('names', []))
    return names


def _choices_options(command: dict) -> dict[str, list[str]]:
    """Return ``{flag: [choice, …]}`` for options with a fixed choices list."""
    result: dict[str, list[str]] = {}
    for opt in command.get('options', []):
        if opt.get('expected_type') == _ET_PROFILE:
            continue
        if opt.get('values') and _nargs_takes_arg(opt.get('arguments')):
            for name in opt.get('names', []):
                result[name] = opt['values']
    return result


def _positional_profile_indices(command: dict) -> list[int]:
    """Return 0-based indices of positionals with ``defined_profile`` type."""
    return [
        i for i, p in enumerate(command.get('positionals', []))
        if p.get('expected_type') == _ET_PROFILE
    ]


def _subcommand_names(command: dict) -> list[str]:
    """Return subcommand names, excluding internal ``__*`` commands."""
    return [
        s['name'] for s in command.get('subcommands', [])
        if not s['name'].startswith('__')
    ]


def _indent(lines: list[str], n: int) -> list[str]:
    pad = '    ' * n
    return [pad + line if line else '' for line in lines]


def _emit_option_arg_block(command: dict, indent_level: int) -> list[str]:
    """Emit a ``case "$prev" in`` block handling option-with-argument slots."""
    profile_pats = _profile_option_patterns(command)
    choices_map = _choices_options(command)

    if not profile_pats and not choices_map:
        return []

    lines: list[str] = ['case "$prev" in']
    if profile_pats:
        pat = '|'.join(profile_pats)
        lines += [
            f'    {pat})',
            f'        {_PROFILE_FN} "$cur"; return ;;',
        ]
    for flag, choices in choices_map.items():
        cw = ' '.join(choices)
        lines += [
            f'    {flag})',
            f'        COMPREPLY=($(compgen -W {repr(cw)} -- "$cur")); return ;;',
        ]
    lines += ['esac']
    return _indent(lines, indent_level)


def _emit_command_body(command: dict, word_index: int,
                       indent_level: int) -> list[str]:
    """Recursively emit the completion body for one command node.

    ``word_index`` is the index into ``$words`` that selects this command
    (1 for root, 2 for first-level subcommands, …).
    """
    lines: list[str] = []
    pad = '    ' * indent_level

    # Option-argument check: $prev is an option that consumes a value.
    lines += _emit_option_arg_block(command, indent_level)

    subs = command.get('subcommands', [])
    opts = _all_option_names(command)
    pos_profile_idx = _positional_profile_indices(command)

    if subs:
        sub_names = _subcommand_names(command)
        visible_subs = [s for s in subs if not s['name'].startswith('__')]
        lines += [f'{pad}case "${{words[{word_index}]}}" in']
        for sub in visible_subs:
            lines += [f'{pad}    {sub["name"]})']
            sub_body = _emit_command_body(sub, word_index + 1, indent_level + 2)
            lines += sub_body
            lines += [f'{pad}        ;;']
        # Default: offer subcommand names and any options at this level.
        word_list = ' '.join(sub_names + opts)
        lines += [
            f'{pad}    *)',
            f'{pad}        COMPREPLY=($(compgen -W {repr(word_list)} -- "$cur"))',
            f'{pad}        ;;',
            f'{pad}esac',
        ]
    else:
        # Leaf command: offer profile positional or option flags.
        if pos_profile_idx:
            lines += [
                f'{pad}# Complete a positional profile name.',
                f'{pad}if [[ $cword -ge {word_index} && "$cur" != -* ]]; then',
                f'{pad}    {_PROFILE_FN} "$cur"; return',
                f'{pad}fi',
            ]
        if opts:
            word_list = ' '.join(opts)
            lines += [
                f'{pad}COMPREPLY=($(compgen -W {repr(word_list)} -- "$cur"))',
            ]

    return lines


def generate_bash(tree: dict) -> str:
    """Return the full static bash completion script for e4s-cl as a string."""

    # Profile-completion helper: embedded python3 script reads the tinydb
    # profile store directly — no dependency on any installed binary.
    profile_fn = textwrap.dedent(f"""\
        {_PROFILE_FN}() {{
            local _profiles
            _profiles=$(python3 - <<'_PY_EOF' 2>/dev/null
        import json, os, pathlib
        _db = (
            pathlib.Path(os.environ.get('__E4S_CL_USER_PREFIX__',
                         str(pathlib.Path.home() / '.local' / 'e4s_cl')))
            / 'user.json'
        )
        try:
            _d = json.loads(open(_db).read())
            print(' '.join(
                v['name']
                for v in _d.get('Profile', {{}}).values()
                if isinstance(v, dict) and 'name' in v
            ))
        except Exception:
            pass
        _PY_EOF
        )
            COMPREPLY=($(compgen -W "$_profiles" -- "$1"))
        }}""")

    root_body = _emit_command_body(tree, word_index=1, indent_level=1)

    main_fn = '\n'.join([
        '_e4s_cl_comp() {',
        '    local cur prev words cword',
        '    # Use bash-completion helper if available; fall back to raw COMP_WORDS.',
        '    _init_completion 2>/dev/null || {',
        '        words=("${COMP_WORDS[@]}")',
        '        cword=$COMP_CWORD',
        '        cur="${COMP_WORDS[$COMP_CWORD]}"',
        '        prev="${COMP_WORDS[$((COMP_CWORD-1))]}"',
        '    }',
        '',
    ] + root_body + ['}'])

    header = textwrap.dedent("""\
        # Bash completion for e4s-cl
        # Auto-generated by scripts/generate_completion_bash.py — do not edit manually.
        # To regenerate: python scripts/generate_completion_bash.py > completions/e4s-cl.bash
        #
        # Installation (pick one):
        #   make install                               — copies to bash-completion user dir
        #   source /path/to/e4s-cl/completions/e4s-cl.bash  — one-time manual setup
        """)

    footer = '\ncomplete -F _e4s_cl_comp -o bashdefault -o default e4s-cl\n'

    return header + '\n' + profile_fn + '\n\n' + main_fn + footer


# ── Completion tree generator — imports e4s_cl ────────────────────────────────

def _generate_completion_tree() -> dict:
    """Build the completion tree by introspecting the live CLI parsers.

    Imports the full ``e4s_cl`` package — call only at build / regeneration
    time, never from a latency-sensitive path.
    """
    from argparse import _StoreAction
    from e4s_cl import cli
    from e4s_cl.cli.commands import __main__ as main_mod  # noqa: F401

    _ARG_TYPES: dict[str, str] = {
        '?': 'ARGS_ATMOSTONE',
        '+': 'ARGS_ATLEASTONE',
        '*': 'ARGS_SOME',
    }

    command_tree = cli._get_commands('e4s_cl.cli.commands')
    command_tree['__module__'] = main_mod

    class _Positional:
        def __init__(self, **kw):
            self.arguments: int | str = 0
            self.values: list[str] = []
            self.expected_type: str = ''
            for k, v in kw.items():
                setattr(self, k, v)

        def json(self) -> dict:
            d: dict = {}
            if self.arguments:
                d['arguments'] = self.arguments
            if self.values:
                d['values'] = self.values
            if self.expected_type:
                d['expected_type'] = self.expected_type
            return d

    class _Option:
        def __init__(self, **kw):
            self.names: list[str] = []
            self.arguments: int | str = 0
            self.values: list[str] = []
            self.expected_type: str = ''
            for k, v in kw.items():
                setattr(self, k, v)

        def json(self) -> dict:
            d: dict = {}
            if self.names:
                d['names'] = self.names
            if self.arguments:
                d['arguments'] = self.arguments
            if self.values:
                d['values'] = self.values
            if self.expected_type:
                d['expected_type'] = self.expected_type
            return d

    class _Command:
        def __init__(self, name: str, dict_: dict):
            self.name = name
            self.subcommands: list = []
            self.positionals: list = []
            self.options: list = []

            # Shallow-copy so dict_.pop('__module__') does not mutate the
            # live e4s_cl.cli._COMMANDS global — which would cause a KeyError
            # on the second call to _generate_completion_tree().
            dict_ = dict(dict_)
            mod = dict_.pop('__module__')
            command = mod.COMMAND

            for action in command.parser.actions:
                nargs = action.nargs
                if isinstance(action, _StoreAction) and not nargs:
                    nargs = 1

                if nargs in _ARG_TYPES:
                    nargs = _ARG_TYPES[nargs]

                if nargs == '...':
                    continue

                if not action.option_strings:
                    if not (nargs and action.type):
                        continue
                    self.positionals.append(_Positional(
                        arguments=nargs,
                        values=sorted(action.choices or []),
                        expected_type=getattr(action.type, '__name__', '') or '',
                    ))
                    continue

                self.options.append(_Option(
                    names=action.option_strings,
                    arguments=nargs,
                    values=sorted(action.choices or []),
                    expected_type=getattr(action.type, '__name__', '') or '',
                ))

            self.subcommands = [_Command(*item) for item in dict_.items()]

        def json(self) -> dict:
            d: dict = {'name': self.name}
            if self.subcommands:
                d['subcommands'] = [c.json() for c in self.subcommands]
            if self.positionals:
                d['positionals'] = [p.json() for p in self.positionals]
            if self.options:
                d['options'] = [o.json() for o in self.options]
            return d

    return _Command('root', command_tree).json()


# ── Public API ────────────────────────────────────────────────────────────────

def regenerate_bash(dest: Path | None = None) -> None:
    """Rebuild the bash completion script from the live CLI parsers.

    Writes to *dest* (defaults to :data:`_BASH_FILE`).  Silently ignores any
    exception so callers are never disrupted.
    """
    target = dest or _BASH_FILE
    try:
        tree = _generate_completion_tree()
        target.parent.mkdir(parents=True, exist_ok=True)
        target.write_text(generate_bash(tree))
    except Exception:  # pragma: no cover
        pass


# ── Developer utilities ──────────────────────────────────────────────────────

def main(argv: list[str] | None = None) -> None:
    """Convenience entry point for regenerating the bash completion script.

    * No arguments → print the path to source for manual setup.
    * ``--regenerate`` → rebuild :data:`_BASH_FILE` from live parsers.

    Users and CI should prefer running the generator directly::

        python scripts/generate_completion_bash.py > completions/e4s-cl.bash
    """
    if argv is None:
        argv = sys.argv[1:]

    if '--regenerate' in argv:
        regenerate_bash()
        return

    if _BASH_FILE.exists():
        print(f'source {_BASH_FILE.resolve()}')
    else:
        print(
            '# e4s-cl completion script not found.\n'
            '# Run  make install  or manually:\n'
            '#   source /path/to/e4s-cl/completions/e4s-cl.bash',
            file=sys.stderr,
        )


if __name__ == '__main__':
    main()




"""
Tests asserting the correct behaviour of the profile create command
"""

import tests
from e4s_cl.model.profile import Profile
from e4s_cl.cli.commands.profile.create import COMMAND as command


class ProfileCreateTest(tests.TestCase):
    """
    The actual tests. Profiles are deleted after every test.
    """

    def setUp(self):
        self.profile_name = 'test_profile'

    def tearDown(self):
        Profile.controller().unselect()
        self.resetStorage()

    def test_create(self):
        self.assertCommandReturnValue(0, command, [self.profile_name])
        self.assertTrue(Profile.controller().one({'name': self.profile_name}))

    def test_create_libraries(self):
        """
        Ensure libraries are added to the profile
        """
        libraries = [f'/tmp/e4s_cl/lib{no}.so' for no in range(5)]

        self.assertCommandReturnValue(
            0, command,
            [self.profile_name, '--libraries', ",".join(libraries)])

        profile = Profile.controller().one({'name': self.profile_name})
        self.assertTrue(profile)

        for libname in libraries:
            self.assertIn(libname, profile.get('libraries', []))

    def test_create_files(self):
        """
        Ensure files are added to the profile
        """
        files = [f'/tmp/e4s_cl/file{no}.txt' for no in range(5)]

        self.assertCommandReturnValue(
            0, command, [self.profile_name, '--files', ",".join(files)])

        profile = Profile.controller().one({'name': self.profile_name})
        self.assertTrue(profile)

        for filename in files:
            self.assertIn(filename, profile.get('files', []))

    def test_create_posix(self):
        """
        Ensure paths are formatted to be posix-compliant
        """
        posix = '/tmp/test.txt'
        non_posix = '/tmp/test.txt///////'

        self.assertCommandReturnValue(
            0, command, [self.profile_name, '--files', non_posix])

        profile = Profile.controller().one({'name': self.profile_name})
        self.assertTrue(profile)

        files = profile.get('files', [])
        self.assertNotIn(non_posix, files)
        self.assertIn(posix, files)

    def test_create_wrong_arguments(self):
        self.assertNotCommandReturnValue(0, command,
                                         [self.profile_name, '--when', 'now'])

        self.assertFalse(Profile.controller().one({'name': self.profile_name}))

    def test_unique(self):
        self.assertCommandReturnValue(0, command, [self.profile_name])
        _, stderr = self.assertNotCommandReturnValue(0, command,
                                                     [self.profile_name])
        self.assertIn('profile create <profile_name>', stderr)
        self.assertIn(
            f"profile create: error: A profile with name='{self.profile_name}' already exists",
            stderr)


fields = {
    ('--image', 'image', 'image:id', 'image_docker_style'),
    ('--image', 'image', '/path/to/image', 'image_singularity_style'),
    ('--backend', 'backend', 'singularity', 'backend'),
    ('--source', 'source', '/path/to/script.sh', 'source'),
    ('--wi4mpi', 'wi4mpi', '/usr/packages/wi4mpi', 'wi4mpi_location'),
}


def wrapper(option, field, argument, test_name):
    """
    Generate tests from a simple pattern to ensure all fields are correctly added
    """

    def generated_test(self):
        self.assertCommandReturnValue(0, command,
                                      [self.profile_name, option, argument])
        profile = Profile.controller().one({'name': self.profile_name})
        self.assertTrue(profile)
        self.assertEqual(argument, profile.get(field))

    generated_test.__name__ = f"test_{test_name}"

    return generated_test


for arguments in fields:
    test = wrapper(*arguments)
    setattr(ProfileCreateTest, test.__name__, test)



"""
Tests for LocalFileStorage (e4s_cl/cf/storage/local_file.py).

These tests exercise the storage layer in isolation using a temporary
directory so they are portable across all environments.  Nothing here
depends on an MPI installation, container runtime, or network access.
"""

import os
import shutil
import tempfile

import tests
from e4s_cl.cf.storage.local_file import LocalFileStorage


class LocalFileStorageCRUDTest(tests.TestCase):
    """Basic create/read/update/delete operations on a temporary storage."""

    def setUp(self):
        self._tmpdir = tempfile.mkdtemp(prefix='e4scl_storage_test_')
        self.storage = LocalFileStorage('test', self._tmpdir)

    def tearDown(self):
        self.storage.disconnect_database()
        shutil.rmtree(self._tmpdir, ignore_errors=True)

    # ------------------------------------------------------------------
    # insert / get / count
    # ------------------------------------------------------------------

    def test_insert_returns_record_with_data(self):
        record = self.storage.insert({'name': 'alpha'}, table_name='T')
        self.assertEqual(record['name'], 'alpha')

    def test_get_by_eid_returns_record(self):
        record = self.storage.insert({'name': 'beta'}, table_name='T')
        found = self.storage.get(record.eid, table_name='T')
        self.assertIsNotNone(found)
        self.assertEqual(found['name'], 'beta')

    def test_get_by_dict_key(self):
        self.storage.insert({'color': 'red'}, table_name='T')
        found = self.storage.get({'color': 'red'}, table_name='T')
        self.assertIsNotNone(found)
        self.assertEqual(found['color'], 'red')

    def test_get_missing_returns_none(self):
        result = self.storage.get({'x': 'absent'}, table_name='T')
        self.assertIsNone(result)

    def test_count_empty(self):
        self.assertEqual(self.storage.count(table_name='T'), 0)

    def test_count_after_inserts(self):
        self.storage.insert({'v': 1}, table_name='T')
        self.storage.insert({'v': 2}, table_name='T')
        self.assertEqual(self.storage.count(table_name='T'), 2)

    # ------------------------------------------------------------------
    # search
    # ------------------------------------------------------------------

    def test_search_all_returns_all_records(self):
        self.storage.insert({'val': 'a'}, table_name='T')
        self.storage.insert({'val': 'b'}, table_name='T')
        results = self.storage.search(table_name='T')
        self.assertEqual(len(results), 2)

    def test_search_with_filter(self):
        self.storage.insert({'color': 'red'}, table_name='T')
        self.storage.insert({'color': 'blue'}, table_name='T')
        results = self.storage.search(keys={'color': 'red'}, table_name='T')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0]['color'], 'red')

    def test_search_no_match_returns_empty(self):
        self.storage.insert({'color': 'red'}, table_name='T')
        results = self.storage.search(keys={'color': 'green'}, table_name='T')
        self.assertEqual(results, [])

    # ------------------------------------------------------------------
    # contains
    # ------------------------------------------------------------------

    def test_contains_true(self):
        self.storage.insert({'id': 'x'}, table_name='T')
        self.assertTrue(self.storage.contains({'id': 'x'}, table_name='T'))

    def test_contains_false(self):
        self.assertFalse(self.storage.contains({'id': 'y'}, table_name='T'))

    def test_contains_none_returns_false(self):
        self.assertFalse(self.storage.contains(None, table_name='T'))

    # ------------------------------------------------------------------
    # update
    # ------------------------------------------------------------------

    def test_update_changes_value(self):
        self.storage.insert({'n': 'old'}, table_name='T')
        self.storage.update({'n': 'new'}, {'n': 'old'}, table_name='T')
        self.assertTrue(self.storage.contains({'n': 'new'}, table_name='T'))
        self.assertFalse(self.storage.contains({'n': 'old'}, table_name='T'))

    # ------------------------------------------------------------------
    # remove / purge
    # ------------------------------------------------------------------

    def test_remove_deletes_matching_records(self):
        self.storage.insert({'kind': 'keep'}, table_name='T')
        self.storage.insert({'kind': 'delete'}, table_name='T')
        self.storage.remove({'kind': 'delete'}, table_name='T')
        self.assertFalse(self.storage.contains({'kind': 'delete'}, table_name='T'))
        self.assertTrue(self.storage.contains({'kind': 'keep'}, table_name='T'))

    def test_purge_clears_all_records(self):
        self.storage.insert({'a': 1}, table_name='T')
        self.storage.insert({'a': 2}, table_name='T')
        self.storage.purge(table_name='T')
        self.assertEqual(self.storage.count(table_name='T'), 0)


class LocalFileStorageKeyValueTest(tests.TestCase):
    """Tests for the dict-like key/value interface."""

    def setUp(self):
        self._tmpdir = tempfile.mkdtemp(prefix='e4scl_kv_test_')
        self.storage = LocalFileStorage('kv', self._tmpdir)

    def tearDown(self):
        self.storage.disconnect_database()
        shutil.rmtree(self._tmpdir, ignore_errors=True)

    def test_setitem_and_getitem(self):
        self.storage['mykey'] = 'myvalue'
        self.assertEqual(self.storage['mykey'], 'myvalue')

    def test_overwrite_existing_key(self):
        self.storage['k'] = 'v1'
        self.storage['k'] = 'v2'
        self.assertEqual(self.storage['k'], 'v2')

    def test_contains_present(self):
        self.storage['present'] = True
        self.assertIn('present', self.storage)

    def test_contains_absent(self):
        self.assertNotIn('absent', self.storage)

    def test_missing_key_raises_key_error(self):
        with self.assertRaises(KeyError):
            _ = self.storage['no_such_key']

    def test_delitem(self):
        self.storage['k'] = 'v'
        del self.storage['k']
        self.assertNotIn('k', self.storage)

    def test_delitem_missing_raises_key_error(self):
        with self.assertRaises(KeyError):
            del self.storage['nope']


class LocalFileStorageFilesystemTest(tests.TestCase):
    """Tests for filesystem helpers."""

    def setUp(self):
        self._tmpdir = tempfile.mkdtemp(prefix='e4scl_fs_test_')
        self.storage = LocalFileStorage('fs', self._tmpdir)

    def tearDown(self):
        self.storage.disconnect_database()
        shutil.rmtree(self._tmpdir, ignore_errors=True)

    def test_is_writable_for_temp_dir(self):
        """A freshly created temp directory must be writable."""
        self.assertTrue(self.storage.is_writable())

    def test_connect_creates_prefix_if_missing(self):
        """connect_filesystem() should create the prefix directory if absent."""
        sub = os.path.join(self._tmpdir, 'new_sub')
        storage = LocalFileStorage('sub', sub)
        self.assertFalse(os.path.isdir(sub))
        storage.connect_filesystem()
        self.assertTrue(os.path.isdir(sub))
        storage.disconnect_database()



"""mpirun launcher support"""

from e4s_cl.cf.launchers import Parser

SCRIPT_NAMES = ['mpirun','mpiexec']

_global = {
    "--allow-run-as-root": 0,
    "--app": 1,
    "--bind-to": 1,
    "--continuous": 0,
    "--daemonize": 0,
    "--debug": 0,
    "--debug-daemons": 0,
    "--debug-daemons-file": 0,
    "--debug-devel": 0,
    "--debug-verbose": 1,
    "--default-hostfile": 1,
    "--disable-recovery": 0,
    "--display-allocation": 0,
    "--display-devel-allocation": 0,
    "--display-devel-map": 0,
    "--display-diffable-map": 0,
    "--display-map": 0,
    "--display-topo": 0,
    "--do-not-launch": 0,
    "--do-not-resolve": 0,
    "--enable-recovery": 0,
    "--forward-signals": 1,
    "--fwd-mpirun-port": 0,
    "--get-stack-traces": 0,
    "--gomca": 2,
    "--gpmixmca": 2,
    "--help": 0,
    "--host": 1,
    "--hostfile": 1,
    "--index-argv-by-rank": 0,
    "--launch-agent": 1,
    "--leave-session-attached": 0,
    "--machinefile": 1,
    "--map-by": 1,
    "--max-restarts": 1,
    "--max-vm-size": 1,
    "--mca": 2,
    "-mca": 2,
    "--merge-stderr-to-stdout": 0,
    "--n": 1,
    "--no-ready-msg": 0,
    "--noprefix": 0,
    "--np": 1,
    "--npernode": 1,
    "--omca": 2,
    "--ompi-server": 1,
    "--output-directory": 1,
    "--output-filename": 1,
    "--output-proctable": 1,
    "--oversubscribe": 0,
    "--parsable": 0,
    "--parseable": 0,
    "--path": 1,
    "--personality": 1,
    "--pmixam": 1,
    "--pmixmca": 2,
    "--prefix": 1,
    "--preload-binary": 0,
    "--preload-files": 1,
    "--prte_info_pretty": 0,
    "--prtemca": 2,
    "--pset": 1,
    "--quiet": 0,
    "--rank-by": 1,
    "--report-bindings": 0,
    "--report-child-jobs-separately": 0,
    "--report-pid": 1,
    "--report-state-on-timeout": 0,
    "--report-uri": 1,
    "--set-cwd-to-session-dir": 0,
    "--set-sid": 0,
    "--show-progress": 0,
    "--stdin": 1,
    "--stop-on-exec": 0,
    "--stream-buffering": 1,
    "--system-server": 0,
    "--tag-output": 0,
    "--test-suicide": 1,
    "--timeout": 1,
    "--timestamp-output": 0,
    "--tmpdir": 1,
    "--tune": 1,
    "--verbose": 0,
    "--version": 0,
    "--wd": 1,
    "--wdir": 1,
    "--xml": 0,
    "--xterm": 1,
    "-H": 1,
    "-N": 1,
    "-V": 0,
    "-c": 1,
    "-configfile": 1,
    "-d": 0,
    "-env": 2,
    "-envall": 0,
    "-envlist": 1,
    "-envnone": 0,
    "-f": 1,
    "-genv": 2,
    "-genvall": 0,
    "-genvlist": 1,
    "-genvnone": 0,
    "-h": 0,
    "-hosts": 1,
    "-n": 1,
    "-np": 1,
    "-q": 0,
    "-s": 0,
    "-v": 0,
    "-wdir": 1,
    "-x": 1,
}

_intel_options = {
    '-aps': 0,
    '-gtool': 0,
    '-gtoolfile': 1,
    '-hosts-group': 1,
    '-mps': 0,
}

_hydra_options = {
    "-bind-to": 1,
    "-disable-auto-cleanup": 0,
    "-disable-hostname-propagation": 0,
    "-disable-x": 0,
    "-enable-x": 0,
    "-errfile-pattern": 0,
    "-iface": 1,
    "-info": 0,
    "-launcher": 1,
    "-launcher-exec": 1,
    "-localhost": 0,
    "-map-by": 1,
    "-membind": 1,
    "-nameserver": 0,
    "-nolocal": 0,
    "-outfile-pattern": 0,
    "-ppn": 1,
    "-prepend-pattern": 0,
    "-prepend-rank": 0,
    "-print-all-exitcodes": 0,
    "-print-rank-map": 0,
    "-rmk": 1,
    "-s": 1,
    "-silent-abort": 0,
    "-tune": 1,
    "-usize": 0,
    "-verbose": 0,
}

_wi4mpi_options = {
    "-E": 1,
    "-extra": 1,
    "--extra": 1,
    "-F": 1,
    "-from": 1,
    "--from": 1,
    "-T": 1,
    "-to": 1,
    "--to": 1,
    "-p": 1,
    "-pm": 1,
    "--pm": 1,
    "-h": 0,
    "-help": 0,
    "--help": 0,
    "-V": 0,
    "-version": 0,
    "--version": 0,
}

_wi4mpi_passed = {
    "-f": 1,
    "-hostfile": 1,
    "--hostfile": 1,
    "-n": 1,
    "-np": 1,
    "-v": 0,
    "-verbose": 0,
    "--verbose": 0,
}

PARSER = Parser({
    **_global,
    **_intel_options,
    **_hydra_options,
    **_wi4mpi_options
})
# Uncomment when py39 comes around
# PARSER = Parser(_global | _intel_options | _hydra_options)



"""
This file contains a BASH script template that will get written in the
container. This script is to be the entry-point as it should set the
execution environment and allows for arbitrary code execution (e.g.
library loading)
"""

from os import chmod, unlink, pathsep
from tempfile import NamedTemporaryFile
from shlex import split
from sotools import is_elf
from e4s_cl import logger
from e4s_cl.error import InternalError

LOGGER = logger.get_logger(__name__)

TEMPLATE = """#!/bin/bash
# Source a user-provided script for convenience
%(source_script)s

# If in debug mode, enable linker debugging from here
%(debugging)s

# Extra environment variables (e.g., Wi4MPI path overrides)
%(extra_env)s

# Enable the host's libraries for this last command by prepending the
# resulting LD_LIBRARY_PATH
export LD_LIBRARY_PATH=%(library_dir)s${LD_LIBRARY_PATH:+:${LD_LIBRARY_PATH}}

# Preload select libraries to ensure they are used, even if RPATHs are set in
# guest libraries.
export LD_PRELOAD=%(preload)s${LD_PRELOAD:+:${LD_PRELOAD}}

# Finally run the command, using the imported linker if applicable
%(linker)s %(command)s
"""


class Entrypoint:
    """
    Objects with exection information that convert to scripts on command
    """

    @property
    def library_dir(self):
        raise NotImplementedError("Deprecated attribute")

    def __init__(self, debug=False):
        self.file_name = None

        # Command to run in the container
        self.command = []

        # Script to source before running anything
        self.source_script_path = ''

        # Directories to add to the runtime linker search path, will take
        # precedence over any paths set before and in the source script
        self.linker_library_path = []

        # List of libraries to preload
        self.preload = []

        # Path to the imported host linker
        self.linker = None

        # Path to the imported host bash/interpreter
        self.interpreter = None

        # Extra environment variables to export (dict of name -> value)
        self.extra_env = {}

        self.debug = debug

    @property
    def command(self):
        return " ".join(self.__command)

    @command.setter
    def command(self, rhs):
        if isinstance(rhs, str):
            self.__command = split(rhs)
        elif isinstance(rhs, list):
            self.__command = rhs
        else:
            raise InternalError(
                f"Invalid format for entrypoint command: {repr(rhs)}")

    @property
    def source_script(self):
        if self.source_script_path:
            return f". {self.source_script_path}"
        return ""

    def __str__(self):
        # Convert to a set to remove duplicates, then as a list to get order
        preload = list(dict.fromkeys(self.preload))

        # The linker statement to prefix to the command
        rtdl = []
        # The command to run
        command = self.command

        # The following is true when the container libc is being replaced
        if self.linker:
            # In case of an ELF binary, start it with the linker; if the
            # command cannot be identified as a binary, run it through bash
            # with the imported linker to ensure the imported libc is loaded
            if self.__command and is_elf(self.__command[0]):
                rtdl = [self.linker]
            elif self.interpreter:
                rtdl = [self.linker, self.interpreter, '-c']
                command = " ".join(['"', *self.__command, '"'])
            else:
                LOGGER.error(
                    "Running a script or command from path is not "
                    "supported in this configuration: missing interpreter"
                    "(linker: %s)", self.linker)

        # Build extra environment exports
        extra_env_lines = []
        for name, value in self.extra_env.items():
            extra_env_lines.append(f'export {name}="{value}"')
        extra_env_str = '\n'.join(extra_env_lines)

        fields = dict(source_script=self.source_script,
                      command=command,
                      library_dir=pathsep.join(
                          map(str, self.linker_library_path)),
                      linker=' '.join(rtdl),
                      preload=':'.join(preload),
                      extra_env=extra_env_str,
                      debugging="export LD_DEBUG=files" if self.debug else '')

        return TEMPLATE % fields

    def setup(self):
        """
        Create a temporary file and print the script in it
        """
        with NamedTemporaryFile('w', delete=False) as script:
            self.file_name = script.name
            script.write(str(self))

        chmod(self.file_name, 0o755)

        sep = "\n" + "".join('=' for _ in range(80))
        LOGGER.debug("Running templated script:%(sep)s\n%(script)s%(sep)s", {
            'sep': sep,
            'script': str(self)
        })

        return self.file_name

    def teardown(self):
        if getattr(self, 'file_name', False):
            unlink(self.file_name)



"""
Tests relating to mpi library version detection
"""

from pathlib import Path
import tests
from e4s_cl.model.profile import Profile
from e4s_cl.cf.libraries import resolve
from e4s_cl.cf.detect_mpi import (
    profile_mpi_name, detect_mpi, _get_mpi_library_version, _suffix_name,
    _extract_mvapich_version, _extract_intel_mpi_version,
    _extract_mpich_version, _extract_cray_mpich_version,
    _extract_open_mpi_version, _get_mpi_handle)

EMPTY_LIB = Path(Path(__file__).parent, 'assets', 'libgver.so.0')


class DetectNameTests(tests.TestCase):
    """
    Tests relating to mpi library version detection
    """

    def tearDown(self):
        Profile.controller().unselect()
        self.resetStorage()
        self.assertFalse(Profile.controller().all())

    def test_suffix(self):
        self.assertEqual(_suffix_name('apple', {}), 'apple')
        self.assertEqual(_suffix_name('apple', {'advanced'}), 'apple')
        self.assertEqual(_suffix_name('apple', {'applesauce'}), 'apple')
        self.assertEqual(_suffix_name('apple', {'advanced', 'apple'}),
                         'apple-2')
        self.assertEqual(_suffix_name('apple', {'advanced', 'apple-4'}),
                         'apple')
        self.assertEqual(
            _suffix_name('apple', {'advanced', 'apple', 'apple-4'}), 'apple-5')

    def test_version_mvapich2(self):
        output_string = """MVAPICH2 Version      :	2.3.5
MVAPICH2 Release date :	Mon November 30 22:00:00 EST 2020
MVAPICH2 Device       :	ch3:mrail
MVAPICH2 configure    :	--prefix=/storage/packages/hackathon-spack/padded-path-length-for-binaries/linux-centos8-x86_64/gcc-8.3.1/mvapich2-2.3.5-7ad2fvw2udf4zchllbq5csvdt437mpw4 --enable-shared --enable-romio --disable-silent-rules --disable-new-dtags --enable-fortran=all --enable-threads=multiple --with-ch3-rank-bits=32 --enable-wrapper-rpath=yes --disable-alloca --enable-fa"""

        self.assertEqual(_extract_mvapich_version(output_string), '2.3.5')

    def test_version_mpich(self):
        output_string = """MPICH Version:	3.3.2
MPICH Release date:	Tue Nov 12 21:23:16 CST 2019
MPICH ABI:	13:8:1
MPICH Device:	ch3:nemesis
MPICH configure:	--prefix=/storage/packages/hackathon-spack/padded-path-length-for-binaries/linux-centos8-x86_64/gcc-8.3.1/mpich-3.3.2-p3cgu65l6tfwgcz2di4c2as3xegyvqwq --disable-silent-rules --enable-shared --with-hwloc-prefix=/storage/packages/hackathon-spack/padded-path-length-for-binaries/linux-centos8-x86_64/gcc-8.3.1/hwloc-2.4.0-t5gpp6x6bw6blyx25ujx4bk5if5iywjt --with-pm=hydra -"""

        self.assertEqual(_extract_mpich_version(output_string), '3.3.2')

    def test_version_cray_mpich(self):
        output_string = """MPI VERSION    : CRAY MPICH version 7.7.14 (ANL base 3.2)
MPI BUILD INFO : Built Tue May 19 13:54:36 2020 (git hash e25eab9) MT-G

"""

        self.assertEqual(_extract_cray_mpich_version(output_string), '7.7.14')

    def test_version_intel(self):
        output_string = """Intel(R) MPI Library 2019 Update 6 for Linux* OS

"""

        self.assertEqual(_extract_intel_mpi_version(output_string),
                         '2019 Update 6')

    def test_version_ompi(self):
        output_string = """Open MPI v4.1.1, package: Open MPI user@host Distribution, ident: 4.1.1, repo rev: v4.1.1, Apr 24, 2021"""

        self.assertEqual(_extract_open_mpi_version(output_string), '4.1.1')

    @tests.skipIf(not resolve('libmpi.so'), "No library to test with")
    def test_extract_handle(self):
        self.assertIsNotNone(_get_mpi_handle(Path(resolve('libmpi.so'))))

    def test_extract_handle_inexistent_path(self):
        self.assertIsNone(_get_mpi_handle(Path('/tmp/sikenolib')))

    def test_extract_handle_bad_path(self):
        self.assertIsNone(_get_mpi_handle(Path('/root')))

    def test_extract_handle_incomplete_library(self):
        self.assertIsNone(_get_mpi_handle(EMPTY_LIB))

    @tests.skipIf(not resolve('libmpi.so'), "No library to test with")
    def test_get_mpi_library_version(self):
        self.assertIsNotNone(
            _get_mpi_library_version(Path(resolve('libmpi.so'))))

        self.assertIsNotNone(
            _get_mpi_library_version(Path(resolve('libmpi.so')).as_posix()))

    def test_get_version_info_inexistent_path(self):
        self.assertFalse(_get_mpi_library_version(Path('/tmp/sikenolib')))

    def test_get_version_info_bad_path(self):
        self.assertFalse(_get_mpi_library_version(Path('/root')))

    def test_get_version_info_incomplete_library(self):
        self.assertFalse(
            _get_mpi_library_version(
                Path(Path(__file__).parent, 'assets', 'libgver.so.0')))

    @tests.skipIf(not resolve('libmpi.so'), "No library to test with")
    def test_detect_mpi(self):
        self.assertTrue(detect_mpi([resolve('libmpi.so')]))

    def test_detect_mpi_no_mpi(self):
        self.assertFalse(detect_mpi([EMPTY_LIB]))

    @tests.skipIf(not resolve('libmpi.so'), "No library to test with")
    def test_rename_profile(self):
        detected = profile_mpi_name([resolve('libmpi.so')])
        self.assertIsNotNone(detected)

    def test_rename_profile_mpiless(self):
        detected = profile_mpi_name([resolve('libc.so.6')])
        self.assertIsNone(detected)



"""Test functions.

Functions used for unit tests of help.py.
"""

import tests
from e4s_cl.cli.commands.help import COMMAND as help_command


class HelpTest(tests.TestCase):
    """Unit tests for `e4s-cl help`"""
    def test_help(self):
        argv = ['profile']
        self.assertCommandReturnValue(0, help_command, argv)

    def test_h_arg(self):
        stdout, _ = self.assertCommandReturnValue(0, help_command, ['-h'])
        self.assertIn('Show this help message and exit', stdout)

    def test_help_arg(self):
        stdout, _ = self.assertCommandReturnValue(0, help_command, ['--help'])
        self.assertIn('Show this help message and exit', stdout)



"""
Tests asserting the correct behaviour of the profile edit command
"""

import tests
from e4s_cl.model.profile import Profile
from e4s_cl.cli.commands.profile.edit import COMMAND


class ProfileEditTest(tests.TestCase):
    """
    The actual tests. Profiles are deleted after every test. More tests are
    added by the code snippet after the class definition.
    """

    def setUp(self):
        self.profile_name = 'test01'
        self.files = ['/tmp/file1']
        self.libraries = ['/tmp/lib1.so']

        Profile.controller().create({
            "name": self.profile_name,
            "image": 'no image',
            "backend": 'bash',
            "source": None,
            "wi4mpi": None,
            "files": self.files,
            "libraries": self.libraries,
        })

    def tearDown(self):
        Profile.controller().unselect()
        self.resetStorage()

    def test_add_file(self):
        filename = '/tmp/file2'
        self.assertCommandReturnValue(
            0, COMMAND, [self.profile_name, '--add-files', filename])

        profile = Profile.controller().one({'name': self.profile_name})
        self.assertTrue(profile)
        self.assertIn(filename, profile.get('files', []))

    def test_add_lib(self):
        filename = '/tmp/lib2.so'
        self.assertCommandReturnValue(
            0, COMMAND, [self.profile_name, '--add-libraries', filename])

        profile = Profile.controller().one({'name': self.profile_name})
        self.assertTrue(profile)
        self.assertIn(filename, profile.get('libraries', []))

    def test_remove_file(self):
        filename = self.files[0]
        self.assertCommandReturnValue(
            0, COMMAND, [self.profile_name, '--remove-files', filename])

        profile = Profile.controller().one({'name': self.profile_name})
        self.assertTrue(profile)
        self.assertNotIn(filename, profile.get('files', []))

    def test_remove_lib(self):
        filename = self.libraries[0]
        self.assertCommandReturnValue(
            0, COMMAND, [self.profile_name, '--remove-libraries', filename])
        profile = Profile.controller().one({'name': self.profile_name})
        self.assertTrue(profile)
        self.assertNotIn(filename, profile.get('libraries', []))

    def test_add_twice_file(self):
        filename = self.files[0]
        _, stderr = self.assertCommandReturnValue(
            0, COMMAND, [self.profile_name, '--add-files', filename])
        self.assertIn(f"File {filename} already in profile's files", stderr)
        filename = f"{self.files[0]}/"
        _, stderr = self.assertCommandReturnValue(
            0, COMMAND, [self.profile_name, '--add-files', filename])
        self.assertIn(f"File {self.files[0]} already in profile's files",
                      stderr)

    def test_add_twice_lib(self):
        filename = self.libraries[0]
        _, stderr = self.assertCommandReturnValue(
            0, COMMAND, [self.profile_name, '--add-libraries', filename])
        self.assertIn(f"File {filename} already in profile's libraries",
                      stderr)
        filename = f"{self.libraries[0]}/"
        _, stderr = self.assertCommandReturnValue(
            0, COMMAND, [self.profile_name, '--add-libraries', filename])
        self.assertIn(
            f"File {self.libraries[0]} already in profile's libraries", stderr)

    def test_remove_nonexistent_file(self):
        filename = "nonexistent"
        _, stderr = self.assertCommandReturnValue(
            0, COMMAND, [self.profile_name, '--remove-files', filename])
        self.assertIn(f"File {filename} not in profile's files", stderr)

    def test_remove_nonexistent_lib(self):
        filename = "nonexistent"
        _, stderr = self.assertCommandReturnValue(
            0, COMMAND, [self.profile_name, '--remove-libraries', filename])
        self.assertIn(f"File {filename} not in profile's libraries", stderr)


fields = {
    ('--image', 'image:id', 'image', 'profile_edit_image'),
    ('--backend', 'singularity', 'backend', 'profile_edit_backend'),
    ('--source', '/tmp/test.sh', 'source', 'profile_edit_source'),
    ('--wi4mpi', '/usr/packages/installation', 'wi4mpi',
     'profile_edit_wi4mpi'),
}


def wrapper(option, value, field, test_name):
    """
    Generate tests from a template and a list of options
    """

    def generated_test(self):
        self.assertCommandReturnValue(0, COMMAND,
                                      [self.profile_name, option, value])

        profile = Profile.controller().one({'name': self.profile_name})
        self.assertTrue(profile)
        self.assertEqual(value, profile.get(field))

    generated_test.__name__ = f"test_{test_name}"

    return generated_test


for arguments in fields:
    test = wrapper(*arguments)
    setattr(ProfileEditTest, test.__name__, test)



"""E4S Container Launcher logging.

E4S CL has two channels for communicating with the user:
    1) sys.stdout via :any:`print`.
       Use this for messages the user has requested, e.g. a project listing.
    2) sys.stdout and sys.stderr via :any:`e4s_cl.logger`.
       Use this for status messages generated by E4S CL.

E4S CL also logs all status messages at the highest reporting level to
a rotating debug file in the user's E4S CL directory, typically "~/.local/e4s_cl".
"""

import os
import re
import sys
import time
import textwrap
import socket
import platform
import string
import logging
import hashlib
import atexit
from pathlib import Path
from time import time
from logging import handlers
from datetime import datetime
from e4s_cl import USER_PREFIX, E4S_CL_VERSION
from e4s_cl.variables import is_parent
from e4s_cl.config import CONFIGURATION

try:
    import termcolor
    COLOR_OUTPUT = True
except ModuleNotFoundError:
    COLOR_OUTPUT = False

# Use isatty to check if the stream supports color
STDOUT_COLOR = os.isatty(sys.stdout.fileno())
STDERR_COLOR = os.isatty(sys.stderr.fileno())

# This is used all over the project, so name translation here
get_logger = logging.getLogger

WARNING = logging.WARNING
ERROR = logging.ERROR


def _prune_ansi(line: str) -> str:
    """
    Remove ANSI color codes from a string
    """
    return re.sub(re.compile('\x1b[^m]+m'), '', line)


def get_terminal_size():
    """Discover the size of the user's terminal.
    
    Several methods are attempted depending on the user's OS.
    If no method succeeds then default to (80, 25).
    
    Returns:
        tuple: (width, height) tuple giving the dimensions of the user's terminal window in characters.
    """
    default_width = 80
    default_height = 25
    dims = _get_term_size_env()

    if not dims:
        dims = _get_term_size_posix()

        if not dims:
            dims = default_width, default_height

    try:
        dims = list(map(int, dims))
    except ValueError:
        dims = default_width, default_height

    width = dims[0] if dims[0] >= 10 else default_width
    height = dims[1] if dims[1] >= 1 else default_height

    return width, height

def _get_term_size_posix():
    """Discover the size of the user's terminal on a POSIX operating system (e.g. Linux).
    
    Returns:
        tuple: (width, height) tuple giving the dimensions of the user's terminal window in characters,
               or None if the size could not be determined.
    """

    # This function follows a POSIX naming scheme, not Python's.
    # pylint: disable=invalid-name
    # Sometimes Pylint thinks termios doesn't exist or doesn't have certain members even when it does.
    # pylint: disable=no-member
    def ioctl_GWINSZ(fd):
        try:
            import fcntl
            import termios
            import struct
            dims = struct.unpack('hh',
                                 fcntl.ioctl(fd, termios.TIOCGWINSZ, '1234'))
        except:  # pylint: disable=bare-except
            return None
        return dims

    dims = ioctl_GWINSZ(0) or ioctl_GWINSZ(1) or ioctl_GWINSZ(2)
    if not dims:
        try:
            fd = os.open(os.ctermid(), os.O_RDONLY)
            dims = ioctl_GWINSZ(fd)
            os.close(fd)
        except:  # pylint: disable=bare-except
            pass
    if not dims:
        return None
    return int(dims[1]), int(dims[0])


def _get_term_size_env():
    """Discover the size of the user's terminal via environment variables.
    
    The user may set the LINES and COLUMNS environment variables to control E4S CL's
    console dimension calculations.
    
    Returns:
        tuple: (width, height) tuple giving the dimensions of the user's terminal window in characters,
               or None if the size could not be determined.
    """
    try:
        return (int(os.environ['COLUMNS']), int(os.environ['LINES']))
    except (KeyError, ValueError):
        return None


def on_stdout(function):

    def wrapper(obj, record):
        text = function(obj, record)
        if STDOUT_COLOR:
            return text
        return _prune_ansi(text)

    return wrapper


def on_stderr(function):

    def wrapper(obj, record):
        text = function(obj, record)
        if STDERR_COLOR:
            return text
        return _prune_ansi(text)

    return wrapper


class LogFormatter(logging.Formatter):
    """Custom log message formatter.
    
    Controls message formatting for all levels.
    
    Args:
        line_width (int): Maximum length of a message line before line is wrapped.
        printable_only (bool): If True, never send unprintable characters to :any:`sys.stdout`.
    """
    # Allow invalid function names to define member functions named after logging levels.
    # pylint: disable=invalid-name

    _printable_chars = set(string.printable)

    def __init__(self, line_width=0, printable_only=False, allow_colors=True):
        super().__init__()
        self.printable_only = printable_only
        self.allow_colors = allow_colors
        self.line_width = line_width

    @on_stderr
    def CRITICAL(self, record):
        return self._colored(self._format_message(record), 'red', None,
                             ['bold'])

    @on_stderr
    def ERROR(self, record):
        return self._colored(self._format_message(record), 'red', None,
                             ['bold'])

    @on_stderr
    def WARNING(self, record):
        return self._colored(self._format_message(record), 'yellow', None,
                             ['bold'])

    @on_stdout
    def INFO(self, record):
        return self._format_message(record, header='[+] ')

    @on_stderr
    def DEBUG(self, record):
        """
        Print a debug message with a neat little header
        """
        message = record.getMessage()
        if self.printable_only and (not set(message).issubset(
                self._printable_chars)):
            message = "<<UNPRINTABLE>>"

        if __debug__:
            marker = self._colored(
                f"[{record.levelname.title()} {record.name}:{record.lineno}]",
                'yellow')
        else:
            marker = self._colored(
                f"[{record.levelname.title()} {getattr(record, 'host', 'localhost')}:{record.process}]",
                'cyan', None, ['bold'])

        return f"{marker} {message}"

    def format(self, record):
        """Formats a log record.
        
        Args:
            record (LogRecord): LogRecord instance to format.
        
        Returns:
            str: The formatted record message.
            
        Raises:
            RuntimeError: No format specified for a the record's logging level.
        """
        try:
            return getattr(self, record.levelname)(record)
        except AttributeError as exc:
            raise RuntimeError(
                f"Unknown record level (name: {record.levelname})") from exc

    def _colored(self, text, *color_args):
        """Insert ANSII color formatting via `termcolor`_.
        
        Text colors:
            * grey
            * red
            * green
            * yellow
            * blue
            * magenta
            * cyan
            * white
        
        Text highlights:
            * on_grey
            * on_red
            * on_green
            * on_yellow
            * on_blue
            * on_magenta
            * on_cyan
            * on_white

        Attributes:
            * bold
            * dark
            * underline
            * blink
            * reverse
            * concealed
        
        .. _termcolor: http://pypi.python.org/pypi/termcolor
        """
        if (COLOR_OUTPUT and self.allow_colors and color_args
                and not sys.stdout.closed):
            return termcolor.colored(text, *color_args)
        return text

    def _format_message(self, record, header=''):
        # Length of the header, pruned from invisible escape characters
        if self.line_width:
            header_length = len(_prune_ansi(header))

            output = []
            text = record.getMessage().split("\n")

            # Strip empty lines at the end only
            while len(text) > 1 and not text[-1]:
                text.pop()

            for line in text:
                output += textwrap.wrap(line,
                                        width=(self.line_width -
                                               header_length))
                if not line:
                    output += ['']

            return textwrap.indent("\n".join(output), header,
                                   lambda line: True)
        return textwrap.indent(record.getMessage().strip(), header,
                               lambda line: True)


def set_log_level(level):
    """Sets :any:`LOG_LEVEL`, the output level for stdout logging objects.
    
    Changes to LOG_LEVEL may affect software package verbosity. 
    
    Args:
        level (str): A string identifying the logging level, e.g. "INFO".
    """
    # Use of global statement is justified in this case.
    # pylint: disable=global-statement
    global LOG_LEVEL
    LOG_LEVEL = level.upper()
    _STDERR_HANDLER.setLevel(LOG_LEVEL) # pylint: disable=E0606


def debug_mode():
    return LOG_LEVEL == 'DEBUG'


LOG_LEVEL = 'INFO'
"""str: The global logging level for stdout loggers and software packages.

Don't change directly. May be changed via :any:`set_log_level`.  
"""

LOG_FILE = Path(USER_PREFIX, 'logs', 'debug_log')
"""str: Absolute path to a log file to receive all debugging output."""

LOG_INDEX = Path(USER_PREFIX, 'logs', 'index.tsv')
"""str: Absolute path to a log file to receive all debugging output."""

LOG_LATEST = Path(USER_PREFIX, 'logs', 'latest')

TERM_SIZE = get_terminal_size()
"""tuple: (width, height) tuple of detected terminal dimensions in characters."""

LINE_WIDTH = TERM_SIZE[0]
"""Width of a line on the terminal.

Uses system specific methods to determine console line width.  If the line
width cannot be determined, the default is 80.
"""

LOG_ID_MARKER = "__E4S_CL_LOG_ID"
"""
Environment variable name: set by the parent for every execution, is used to
group debug logs in folders
"""


def setup_process_logger(name: str) -> logging.Logger:
    """
    Create and setup handlers of a Logger object meant to log errors of
    a subprocess
    """

    # Locate and ensure the log file directory is writeable
    # - Compatible with symlinks in LOG_ID
    log_file = Path(_LOG_FILE_PREFIX, LOG_ID, name).resolve() # pylint: disable=E0606
    Path.mkdir(log_file.parent, parents=True, exist_ok=True)

    # Create a logger in debug mode
    process_logger = logging.getLogger(name)
    if CONFIGURATION.disable_ranked_log:
        process_logger.setLevel(logging.ERROR)
        process_logger.propagate = False
    else:
        process_logger.setLevel(logging.DEBUG)

        # Log process data to file
        handler = logging.FileHandler(log_file,
                                      mode='a',
                                      encoding='utf-8',
                                      delay=True)
        handler.setFormatter(LogFormatter(line_width=120, allow_colors=False))
        process_logger.addHandler(handler)

        # This disables the propagation along the logger tree, to avoid getting
        # everything on stderr
        process_logger.propagate = False

    return process_logger


def is_available(file: Path) -> bool:
    """
    Returns True if the path passed as an argument is accessible to write
    The parents of the passed file will be create if not existent
    """
    parent_dir = file.parent

    if parent_dir.is_symlink():
        parent_dir = parent_dir.resolve()

    try:
        # Ensure the file directory is accessible, create it if need be
        Path.mkdir(parent_dir, parents=True, exist_ok=True)

        with open(file, 'a', encoding='utf-8') as _:
            pass
    except OSError as exc:
        _ROOT_LOGGER.debug("Failed to open file %s: %s", file.as_posix(),
                           exc.strerror)
        return False
    return True


def add_file_handler(
    log_file: Path,
    logger: logging.Logger,
    formatter: logging.Formatter = LogFormatter(line_width=120,
                                                allow_colors=False)
) -> bool:
    """
    Add a file handler to a Logger object
    """
    if not is_available(log_file):
        return False

    file_handler = handlers.TimedRotatingFileHandler(log_file,
                                                     when='D',
                                                     interval=1,
                                                     backupCount=3)
    file_handler.setFormatter(formatter)
    file_handler.setLevel(logging.DEBUG)
    logger.addHandler(file_handler)

    return True


def update_symlink_latest():
    if Path(LOG_FILE.parent, LOG_ID).exists():
        try:
            LOG_LATEST.unlink(missing_ok=True)
            LOG_LATEST.symlink_to(Path(LOG_FILE.parent, LOG_ID))
        except OSError as err:
            _ROOT_LOGGER.debug("Updating symlink %s failed: %s",
                               LOG_LATEST.as_posix(), str(err))


_ROOT_LOGGER = logging.getLogger()
if not _ROOT_LOGGER.handlers:
    _ROOT_LOGGER.setLevel(logging.DEBUG)
    _LOG_FILE_PREFIX = LOG_FILE.parent

    # Setup output on stderr
    _STDERR_HANDLER = logging.StreamHandler(sys.stderr)
    _STDERR_HANDLER.setFormatter(LogFormatter(printable_only=False))
    _STDERR_HANDLER.setLevel(LOG_LEVEL)

    _ROOT_LOGGER.addHandler(_STDERR_HANDLER)

# Create a hash for the execution ID, and dictate to dump logs in the
# corresponding folder
if is_parent():
    grinder = hashlib.sha256()
    grinder.update(str(time()).encode())

    LOG_ID = grinder.hexdigest()
    os.environ[LOG_ID_MARKER] = LOG_ID
else:
    LOG_ID = os.environ.get(LOG_ID_MARKER, 'NOID')

if is_parent():
    # Add a file handler, location depending on the status of the process
    add_file_handler(LOG_FILE, _ROOT_LOGGER)

    # When running as e4s-cl
    if Path(sys.argv[0]).name == 'e4s-cl':
        # Log the command in a database to ease human lookup
        index_logger = logging.getLogger("index")
        index_logger.propagate = False
        add_file_handler(LOG_INDEX,
                         index_logger,
                         formatter=logging.Formatter(
                             fmt=f"%(asctime)s\t{LOG_ID}\t%(message)s"))
        index_logger.info(" ".join(sys.argv))

        # Registers the update of the latest logs symlink to be run
        # at the end of the execution
        atexit.register(update_symlink_latest)

    # pylint: disable=logging-not-lazy
    _ROOT_LOGGER.debug(
        ("\n%(bar)s\n"
         "E4S CONTAINER LAUNCHER LOGGING INITIALIZED\n"
         "\n"
         "Timestamp         : %(timestamp)s\n"
         "Hostname          : %(hostname)s\n"
         "Platform          : %(platform)s\n"
         "Version           : %(version)s\n"
         "Python Version    : %(pyversion)s\n"
         "Working Directory : %(cwd)s\n"
         "Terminal Size     : %(termsize)s\n"
         "Frozen            : %(frozen)s\n"
         "Log ID            : %(logid)s\n"
         "%(bar)s\n") % {
             'bar': '#' * LINE_WIDTH,
             'timestamp': str(datetime.now()),
             'hostname': socket.gethostname(),
             'platform': platform.platform(),
             'version': E4S_CL_VERSION,
             'pyversion': platform.python_version(),
             'cwd': os.getcwd(),
             'termsize': 'x'.join([str(_) for _ in TERM_SIZE]),
             'frozen': getattr(sys, 'frozen', False),
             'logid': LOG_ID,
         })
elif not CONFIGURATION.disable_ranked_log:
    _log_file = Path(_LOG_FILE_PREFIX, LOG_ID, f"e4s_cl.{os.getpid()}")
    add_file_handler(_log_file, _ROOT_LOGGER)



"""
Model definition of the MVC architecture
"""

from e4s_cl import logger
from e4s_cl.error import IncompatibleRecordError, ModelError, InternalError
from e4s_cl.cf.storage import StorageRecord
from e4s_cl.mvc.controller import Controller

LOGGER = logger.get_logger(__name__)


class ModelMeta(type):
    """Constructs model attributes, configures defaults, and establishes relationships."""

    def __new__(mcs, name, bases, dct):
        if dct['__module__'] != __name__:
            # Each Model subclass has its own relationships
            dct.update({'associations': {}, 'references': set()})
            # The default model name is the class name
            if dct.get('name', None) is None:
                dct['name'] = name
            # Model subclasses must define __attributes__ as a callable.
            # We make the callable a staticmethod to prevent method binding.
            try:
                dct['__attributes__'] = staticmethod(dct['__attributes__'])
            except KeyError as err:
                raise InternalError(
                    f"Model class {name} does not define '__attributes__'"
                ) from err
            # Replace attributes with a callable property (defined below).  This is to guarantee
            # that model attributes won't be constructed until all Model subclasses have been constructed.
            dct['attributes'] = ModelMeta.attributes
            # Replace key_attribute with a callable property (defined below). This is to set
            # the key_attribute member after the model attributes have been constructed.
            dct['key_attribute'] = ModelMeta.key_attribute
        return type.__new__(mcs, name, bases, dct)

    @property
    def attributes(cls):
        # pylint: disable=attribute-defined-outside-init
        try:
            return cls._attributes
        except AttributeError:
            cls._attributes = cls.__attributes__()
            cls._construct_relationships()
            return cls._attributes

    @property
    def key_attribute(cls):
        """
        Get the key attribute of a class
        """
        # pylint: disable=no-value-for-parameter
        try:
            return cls._key_attribute
        except AttributeError as err:
            for attr, props in cls.attributes.items():
                if 'primary_key' in props:
                    cls._key_attribute = attr
                    break
            else:
                raise ModelError(
                    cls,
                    "No attribute has the 'primary_key' property set to 'True'"
                ) from err
            return cls._key_attribute


class Model(StorageRecord, metaclass=ModelMeta):
    """The "M" in `MVC`_.

    Attributes:
        name (str): Name of the model.
        associations (dict): (Controller, str) tuples keyed by attribute name defining attribute associations.
        references (set): (Controller, str) tuples listing foreign models referencing this model.  
        attributes (dict): Model attributes.
        key_attribute (str): Name of an attribute that serves as a unique identifier. 
        
    .. _MVC: https://en.wikipedia.org/wiki/Model-view-controller
    """

    __controller__ = Controller
    __attributes__ = NotImplemented

    name = None
    associations = {}
    references = set()
    attributes = {}
    key_attribute = None

    def __init__(self, record):
        deprecated = [attr for attr in record if attr not in self.attributes]
        if deprecated:
            try:
                title = f"{self.name} '{record[self.key_attribute]}'"
            except (KeyError, ModelError):
                title = f"{self.name}"
            LOGGER.debug("Ignorning deprecated attributes %s in %s",
                         deprecated, title)
        super().__init__(
            record.storage, record.eid,
            (item for item in record.items() if item[0] not in deprecated))

    def __setitem__(self, key, value):
        raise InternalError(
            "Use controller(storage).update() to alter records")

    def __delitem__(self, key):
        raise InternalError(
            "Use controller(storage).update() to alter records")

    def get_or_default(self, key):
        """Returns the attribute's value or default value.
        
        Args:
            key (str): Attribute name.
        
        Returns:
            If the attribute is set then the attribute's value is returned.
            If the attribute is not set then the attribute's default value is returned.

        Raises:
            If the attribute is not set and has no default value then a KeyError is raised.
        """
        try:
            return self[key]
        except KeyError:
            return self.attributes[key]['default']

    def on_create(self):
        """Callback to be invoked after a new data record is created."""

    def on_update(self, changes):
        """Callback to be invoked after a data record is updated."""

    def on_delete(self):
        """Callback to be invoked before a data record is deleted."""

    @classmethod
    def controller(cls, storage):
        return cls.__controller__(cls, storage)

    @classmethod
    def _construct_relationships(cls):
        primary_key = None
        for attr, props in cls.attributes.items():
            model_attr_name = cls.name + "." + attr
            if 'collection' in props and 'via' not in props:
                raise ModelError(
                    cls,
                    f"{model_attr_name}: collection does not define 'via'")
            if 'via' in props and not ('collection' in props
                                       or 'model' in props):
                raise ModelError(
                    cls,
                    f"{model_attr_name}: defines 'via' property but not 'model'"
                    "or 'collection'")
            if not isinstance(props.get('unique', False), bool):
                raise ModelError(
                    cls, f"{model_attr_name}: invalid value for 'unique'")
            if not isinstance(props.get('description', ''), str):
                raise ModelError(
                    cls, f"{model_attr_name}: invalid value for 'description'")
            if props.get('primary_key', False):
                if primary_key is not None:
                    raise ModelError(
                        cls,
                        f"{model_attr_name}: primary key previously specified "
                        f"as {primary_key}")
                primary_key = attr

            via = props.get('via', None)
            foreign_cls = props.get('model', props.get('collection', None))
            if not foreign_cls:
                continue
            try:
                if not issubclass(foreign_cls, Model):
                    raise TypeError
            except TypeError as err:
                raise ModelError(
                    cls,
                    f"{model_attr_name}: Invalid foreign model controller: "
                    f"{foreign_cls}") from err

            forward = (foreign_cls, via)
            reverse = (cls, attr)
            if not via:
                foreign_cls.references.add(reverse)
            else:
                foreign_cls.associations[via] = reverse
                foreign_model_attr_name = foreign_cls.name + "." + via
                try:
                    via_props = foreign_cls.attributes[via]
                except KeyError as err:
                    raise ModelError(
                        cls, f"{model_attr_name}: 'via' references undefined "
                        f"attribute '{foreign_model_attr_name}'") from err
                via_attr_model = via_props.get(
                    'model', via_props.get('collection', None))
                if not via_attr_model:
                    raise ModelError(
                        cls,
                        f"{model_attr_name}: 'via' on non-model attribute "
                        f"'{foreign_model_attr_name}'")
                if via_attr_model is not cls:
                    raise ModelError(
                        cls,
                        f"Attribute '{foreign_model_attr_name}' referenced by "
                        f"'via' in '{attr}' does not define 'collection' or "
                        f"'model' of type '{cls.name}'")
                try:
                    existing = cls.associations[attr]
                except KeyError:
                    cls.associations[attr] = forward
                else:
                    if existing != forward:
                        raise ModelError(
                            cls,
                            f"{model_attr_name}: conflicting associations: "
                            f"'{existing}' vs. '{forward}'")

    @classmethod
    def validate(cls, data):
        """Validates data against the model.
        
        Args:
            data (dict): Data to validate, may be None.

        Returns:
            dict: Validated data or None if `data` is None.
            
        Raises:
            ModelError: The given data doesn't fit the model.
        """
        if data is None:
            return None
        for key in data:
            if key not in cls.attributes:
                raise ModelError(cls, f"no attribute named '{key}'")
        validated = {}
        for attr, props in cls.attributes.items():
            # Check required fields and defaults
            try:
                validated[attr] = data[attr]
            except KeyError as err:
                if 'required' in props:
                    if props['required']:
                        raise ModelError(
                            cls, f"'{attr}' is required but was not defined"
                        ) from err
                elif 'default' in props:
                    validated[attr] = props['default']
            # Check collections
            if 'collection' in props:
                value = data.get(attr, [])
                if not value:
                    value = []
                elif not isinstance(value, list):
                    raise ModelError(
                        cls,
                        f"Value supplied for '{attr}' is not a list: {value}")
                else:
                    for eid in value:
                        try:
                            int(eid)
                        except ValueError as err:
                            raise ModelError(
                                cls,
                                f"Invalid non-integer ID '{eid}' in '{attr}'"
                            ) from err
                validated[attr] = value
            # Check model associations
            elif 'model' in props:
                value = data.get(attr, None)
                if value is not None:
                    try:
                        if int(value) != value:
                            raise ValueError
                    except ValueError as err:
                        raise ModelError(
                            cls,
                            f"Invalid non-integer ID '{value}' in '{attr}'"
                        ) from err
                    validated[attr] = value
        return validated

    @classmethod
    def construct_condition(cls,
                            args,
                            attr_defined=None,
                            attr_undefined=None,
                            attr_eq=None,
                            attr_ne=None):
        """Constructs a compatibility condition, see :any:`check_compatibility`.
        
        The returned condition is a callable that accepts four arguments:
            * lhs (Model): The left-hand side of the `check_compatibility` operation.
            * lhs_attr (str): Name of the attribute that defines the 'compat' property.
            * lhs_value: The value of the attribute that defines the 'compat' property.
            * rhs (Model): Controller of the data record we are checking against.
        
        The `condition` callable raises a :any:`IncompatibleRecordError` if the compared attributes 
        are fatally incompatibile, i.e. the user's operation is guaranteed to fail with the chosen 
        records. It may emit log messages to indicate that the records are not perfectly compatible 
        but that the user's operation is still likely to succeed with the chosen records.
        
        See :any:`require`, :any:`encourage`, :any:`discourage`, :any:`exclude` for common conditions.
        
        args[0] specifies a model attribute to check.  If args[1] is given, it is a value to compare 
        the specified attribute against or a callback function.  If args[1] is callable, it must check attribute 
        existance and value correctness and throw the appropriate exception and/or emit log messages.
        
        Args:
            args (tuple): Attribute name in args[0] and, optionally, attribute value in args[1].
            attr_defined: Callback function to be invoked when the attribute is defined.
            attr_undefined: Callback function to be invoked when the attribute is undefined.
            attr_eq: Callback function to be invoked when the attribute is equal to args[1].
            attr_ne: Callback function to be invoked when the attribute is not equal to args[1].

        Returns:
            Callable condition object for use with :any:`check_compatibility`.
        """
        rhs_attr = args[0]
        try:
            checked_value = args[1]
        except IndexError:

            def condition(lhs, lhs_attr, lhs_value, rhs):
                if isinstance(rhs, cls):
                    if rhs_attr in rhs:
                        if attr_defined:
                            attr_defined(lhs, lhs_attr, lhs_value, rhs,
                                         rhs_attr)
                    else:
                        if attr_undefined:
                            attr_undefined(lhs, lhs_attr, lhs_value, rhs,
                                           rhs_attr)
        else:
            if callable(checked_value):

                def condition(lhs, lhs_attr, lhs_value, rhs):
                    if isinstance(rhs, cls):
                        checked_value(lhs, lhs_attr, lhs_value, rhs, rhs_attr)
            else:

                def condition(lhs, lhs_attr, lhs_value, rhs):
                    if isinstance(rhs, cls):
                        try:
                            rhs_value = rhs[rhs_attr]
                        except KeyError:
                            if attr_undefined:
                                attr_undefined(lhs, lhs_attr, lhs_value, rhs,
                                               rhs_attr)
                        else:
                            if attr_eq:
                                if rhs_value == checked_value:
                                    attr_eq(lhs, lhs_attr, lhs_value, rhs,
                                            rhs_attr, checked_value)
                            elif attr_ne:
                                if rhs_value != checked_value:
                                    attr_ne(lhs, lhs_attr, lhs_value, rhs,
                                            rhs_attr, checked_value)
                            elif attr_defined:
                                attr_defined(lhs, lhs_attr, lhs_value, rhs,
                                             rhs_attr)

        return condition

    @classmethod
    def require(cls, *args):
        """Constructs a compatibility condition to enforce required conditions.
        
        The condition will raise a :any:`IncompatibleRecordError` if the specified attribute is
        undefined or not equal to the specified value (if given).
        
        Args:
            *args: Corresponds to `args` in :any:`construct_condition`. 
        
        Returns:
            Callable condition object for use with :any:`check_compatibility`
            
        Examples:
            'have_cheese' must be True::
            
                CheeseShop.require('have_cheese', True)
             
            'have_cheese' must be set to any value::
                
                CheeseShop.require('have_cheese')
            
            The value of 'have_cheese' will be checked for correctness by 'cheese_callback'::
            
                CheeseShop.require('have_cheese', cheese_callback)
        """

        def attr_undefined(lhs, lhs_attr, lhs_value, rhs, rhs_attr):
            lhs_name = lhs.name.lower()
            rhs_name = rhs.name.lower()
            raise IncompatibleRecordError(
                f"{lhs_attr} = {lhs_value} in {lhs_name} requires {rhs_attr} "
                f"be defined in {rhs_name} but it is undefined")

        def attr_ne(lhs, lhs_attr, lhs_value, rhs, rhs_attr, checked_value):
            lhs_name = lhs.name.lower()
            rhs_name = rhs.name.lower()
            rhs_value = rhs[rhs_attr]
            raise IncompatibleRecordError(
                f"{lhs_attr} = {lhs_value} in {lhs_name} requires {rhs_attr} "
                f"= {checked_value} in {rhs_name} but it is {rhs_value}")

        return cls.construct_condition(args,
                                       attr_undefined=attr_undefined,
                                       attr_ne=attr_ne)

    @classmethod
    def encourage(cls, *args):
        """Constructs a compatibility condition to make recommendations.
        
        The condition will emit warnings messages if the specified attribute is
        undefined or not equal to the specified value (if given).
        
        Args:
            *args: Corresponds to `args` in :any:`construct_condition`. 
        
        Returns:
            Callable condition object for use with :any:`check_compatibility`
            
        Examples:
            'have_cheese' should be True::
            
                CheeseShop.encourage('have_cheese', True)
             
            'have_cheese' should be set to any value::
                
                CheeseShop.encourage('have_cheese')
            
            The value of 'have_cheese' will be checked for correctness by 'cheese_callback'::
            
                CheeseShop.encourage('have_cheese', cheese_callback)
        """

        def attr_undefined(lhs, lhs_attr, lhs_value, rhs, rhs_attr):
            lhs_name = lhs.name.lower()
            rhs_name = rhs.name.lower()
            LOGGER.warning(
                "%s = %s in %s recommends %s be defined in %s but it is undefined",
                lhs_attr, lhs_value, lhs_name, rhs_attr, rhs_name)

        def attr_ne(lhs, lhs_attr, lhs_value, rhs, rhs_attr, checked_value):
            lhs_name = lhs.name.lower()
            rhs_name = rhs.name.lower()
            rhs_value = rhs[rhs_attr]
            LOGGER.warning(
                "%s = %s in %s recommends %s = %s in %s but it is %s",
                lhs_attr, lhs_value, lhs_name, rhs_attr, checked_value,
                rhs_name, rhs_value)

        return cls.construct_condition(args,
                                       attr_undefined=attr_undefined,
                                       attr_ne=attr_ne)

    @classmethod
    def discourage(cls, *args):
        """Constructs a compatibility condition to make recommendations.
        
        The condition will emit warnings messages if the specified attribute is
        defined or equal to the specified value (if given).
        
        Args:
            *args: Corresponds to `args` in :any:`construct_condition`. 
        
        Returns:
            Callable condition object for use with :any:`check_compatibility`
            
        Examples:
            'have_cheese' should not be True::
            
                CheeseShop.discourage('have_cheese', True)
             
            'have_cheese' should not be set to any value::
                
                CheeseShop.discourage('have_cheese')
            
            The value of 'have_cheese' will be checked for correctness by 'cheese_callback'::
            
                CheeseShop.discourage('have_cheese', cheese_callback)
        """

        def attr_defined(lhs, lhs_attr, lhs_value, rhs, rhs_attr):
            lhs_name = lhs.name.lower()
            rhs_name = rhs.name.lower()
            LOGGER.warning("%s = %s in %s recommends %s be undefined in %s",
                           lhs_attr, lhs_value, lhs_name, rhs_attr, rhs_name)

        def attr_eq(lhs, lhs_attr, lhs_value, rhs, rhs_attr, checked_value):
            lhs_name = lhs.name.lower()
            rhs_name = rhs.name.lower()
            LOGGER.warning("%s = %s in %s recommends against %s = %s in %s",
                           lhs_attr, lhs_value, lhs_name, rhs_attr,
                           checked_value, rhs_name)

        return cls.construct_condition(args,
                                       attr_defined=attr_defined,
                                       attr_eq=attr_eq)

    @classmethod
    def exclude(cls, *args):
        """Constructs a compatibility condition to enforce required conditions.
        
        The condition will raise a :any:`IncompatibleRecordError` if the specified attribute is
        defined or equal to the specified value (if given).
        
        Args:
            *args: Corresponds to `args` in :any:`construct_condition`. 
        
        Returns:
            Callable condition object for use with :any:`check_compatibility`
            
        Examples:
            'yellow_cheese' must not be 'American'::
            
                CheeseShop.exclude('yellow_cheese', 'American')
             
            'blue_cheese' must not be set to any value::
                
                CheeseShop.exclude('blue_cheese')
            
            The value of 'have_cheese' will be checked for correctness by 'cheese_callback'::
            
                CheeseShop.exclude('have_cheese', cheese_callback)
        """

        def attr_defined(lhs, lhs_attr, lhs_value, rhs, rhs_attr):
            lhs_name = lhs.name.lower()
            rhs_name = rhs.name.lower()
            raise IncompatibleRecordError(
                f"{lhs_attr} = {lhs_value} in {lhs_name} requires {rhs_attr}"
                f"be undefined in {rhs_name}")

        def attr_eq(lhs, lhs_attr, lhs_value, rhs, rhs_attr, checked_value):
            lhs_name = lhs.name.lower()
            rhs_name = rhs.name.lower()
            raise IncompatibleRecordError(
                f"{lhs_attr} = {lhs_value} in {lhs_name} is incompatible with "
                f"{rhs_attr} = {checked_value} in {rhs_name}")

        return cls.construct_condition(args,
                                       attr_defined=attr_defined,
                                       attr_eq=attr_eq)

    def check_compatibility(self, rhs):
        """Test this record for compatibility with another record.
        
        Operations combining data from multiple records must know that the records are mutually 
        compatible.  This routine checks the 'compat' property of each attribute (if set) to enforce
        compatibility.  'compat' is a dictionary where the keys are values or callables and the 
        values are tuples of compatibility conditions.  If the attribute with the 'compat' property
        is one of the key values then the conditions are checked.  The general form of 'compat' is::
        
            {
            (value|callable): (condition, [condition, ...]),
            [(value|callable): (condition, [condition, ...]), ...]
            }
            
        Use tuples to join multiple conditions.  If only one condition is needed then you do
        not need to use a tuple.
        
        'value' may either be a literal value (e.g. True or "oranges") or a callable accepting
        one argument and returning either True or False.  The attribute's value is passed to the
        callable to determine if the listed conditions should be checked.  If 'value' is a literal
        then the listed conditions are checked when the attribute's value matches 'value'.     
        
        See :any:`require`, :any:`encourage`, :any:`discourage`, :any:`exclude` for common conditions.
        
        Args:
            rhs (Model): Check compatibility with this data record.
            
        Raises:
            IncompatibleRecordError: The two records are not compatible.

        Examples:

            Suppose we have this data:
            ::
            
                Programmer.attributes = {
                    'hungry': {
                        'type': 'boolean',
                        'compat': {True: (CheeseShop.require('have_cheese', True),
                                          CheeseShop.encourage('chedder', 'Wisconsin'),
                                          ProgramManager.discourage('holding_long_meeting', True),
                                          Roommate.exclude('steals_food', True)}
                    }
                }

                bob = Programmer({'hungry': True})
                world_o_cheese = CheeseShop({'have_cheese': False, 'chedder': 'Wisconsin'})
                cheese_wizzard = CheeseShop({'have_cheese': True, 'chedder': 'California'})
                louis = ProgramManager({'holding_long_meeting': True})
                keith = Roommate({'steals_food': True})
                
            These expressions raise :any:`IncompatibleRecordError`::
            
                bob.check_compatibility(world_o_cheese)   # Because have_cheese == False
                bob.check_compatibility(keith)            # Because steals_food == True
            
            These expressions generate warning messages::
            
                bob.check_compatibility(cheese_wizzard)   # Because chedder != Wisconsin
                bob.check_compatibility(louis)            # Because holding_long_meeting == True
                
            If ``bob['hungry'] == False`` or if the 'hungry' attribute were not set then all 
            the above expressions do nothing.
        """
        as_tuple = lambda x: x if isinstance(x, tuple) else (x, )
        for attr, props in self.attributes.items():
            try:
                compat = props['compat']
            except KeyError:
                continue
            try:
                attr_value = self[attr]
            except KeyError:
                continue
            for value, conditions in compat.items():
                if (callable(value)
                        and value(attr_value)) or attr_value == value:
                    for condition in as_tuple(conditions):
                        condition(self, attr, attr_value, rhs)

    @classmethod
    def filter_arguments(cls, args):
        from e4s_cl.cli.arguments import ArgumentsNamespace
        filtered = dict(item for item in vars(args).items()
                        if item[0] in cls.attributes)
        return ArgumentsNamespace(**filtered)



"""
Tests ensuring the init command behaves as intented
"""

import os
from itertools import combinations
import tests
from tests import TEST_SYSTEM
from e4s_cl.util import which
from e4s_cl.model.profile import Profile
from e4s_cl.cf.libraries import resolve
from e4s_cl.cli.commands.init import COMMAND


class InitTest(tests.TestCase):
    """
    Partial class definition: more tests are defined below
    """

    def tearDown(self):
        Profile.controller().unselect()
        self.resetStorage()


options = [
    ('--mpi', '/path/to/installation'),
    ('--launcher', '/path/to/binary'),
    ('--launcher_args', "'-np 8192'"),
    ('--wi4mpi', "/tmp"),
]



"""
Dummy container used during tests
"""

from typing import List
from e4s_cl.cf.containers import Container

DEBUG_BACKEND = True
NAME = 'dummy'
EXECUTABLES = ['bash']
MIMES = []


class DummyContainer(Container):

    def run(self, command: List[str], overload: bool = True) -> int:
        pass


CLASS = DummyContainer



"""
Definition of storage levels
"""

from e4s_cl import SYSTEM_PREFIX, USER_PREFIX
from e4s_cl.cf.storage import StorageError
from e4s_cl.cf.storage.local_file import LocalFileStorage

SYSTEM_STORAGE = LocalFileStorage('system', SYSTEM_PREFIX)
"""System-level data storage."""

USER_STORAGE = LocalFileStorage('user', USER_PREFIX)
"""User-level data storage."""

ORDERED_LEVELS = (USER_STORAGE, SYSTEM_STORAGE)
"""All storage levels in their preferred order."""

STORAGE_LEVELS = {level.name: level for level in ORDERED_LEVELS}
"""All storage levels indexed by their names."""


def highest_writable_storage():
    """
    Return the first storage location in order that can be accessed to write in
    """
    try:
        return highest_writable_storage.value
    except AttributeError as err:
        for storage in reversed(ORDERED_LEVELS):
            if storage.is_writable():
                highest_writable_storage.value = storage
                break
        else:
            raise StorageError("No writable storage levels") from err
        return highest_writable_storage.value



"""
Assert basic configuration capabilities are supported
"""

import tests
import shlex
from io import StringIO
from pathlib import Path
from unittest.mock import patch
import e4s_cl.config
from e4s_cl.config import (
    ALLOWED_CONFIG,
    Configuration,
    ConfigurationError,
    flatten,
)
from e4s_cl.variables import set_dry_run
from e4s_cl.cf.containers import Container
from e4s_cl.cli.commands.launch import COMMAND as launch_command


class ConfigTest(tests.TestCase):
    """Unit tests for e4s-cl's configuration file use"""

    def test_flat(self):
        data = {
            'container_directory': '/diffdirectory',
            'launcher_options': '-n 8',
            'singularity': {
                'options': '--hostname diffname',
                'build_options': {
                    'location': '-s'
                }
            }
        }
        flat_data = {
            'container_directory': '/diffdirectory',
            'launcher_options': '-n 8',
            'singularity_options': '--hostname diffname',
            'singularity_build_options_location': '-s'
        }
        self.assertEqual(flatten(data), flat_data)

    def test_bad_type(self):
        data = """---
launcher_options: 8"""

        with self.assertRaises(ConfigurationError):
            config = Configuration.create_from_string(data)

    def test_merge(self):
        fields1 = {'a': 1, 'b': 3}
        fields2 = {'a': 5}
        fields3 = {'c': 0, 'd': 0, 'e': 0}
        c1, c2, c3 = Configuration(fields1), Configuration(
            fields2), Configuration(fields3)

        merged = c1 | c2 | c3
        expected = Configuration(dict(a=5, b=3, c=0, d=0, e=0))

        self.assertEqual(merged._fields, expected._fields)

    def test_completion(self):
        c = Configuration.default()

        for field in ALLOWED_CONFIG.flatten():
            self.assertEqual(getattr(c, field.key, None), field.default())

    def test_access(self):
        fields = {'a': 1, 'b': 2}
        c = Configuration(fields)
        self.assertEqual(getattr(c, 'a', None), 1)
        self.assertEqual(getattr(c, 'b', None), 2)
        self.assertIsNone(getattr(c, 'c', None))

    def test_assets(self):
        self.assertNotEqual(
            Configuration.create_from_file(tests.ASSETS / "e4s-cl.yaml"),
            Configuration.default())



"""
Direct unit tests for the MVC Controller layer (e4s_cl/mvc/controller.py).

Rather than testing through the CLI, these tests drive Controller methods
directly via the Profile model and the test-configured USER_STORAGE so that
each method is exercised in isolation.  setUp/tearDown reset storage before
and after every case to keep tests independent.
"""

import tests
from e4s_cl.model.profile import Profile
from e4s_cl.error import UniqueAttributeError, ModelError


class ControllerCRUDTest(tests.TestCase):
    """create / one / all / count / exists / delete."""

    def setUp(self):
        self.resetStorage()
        self._ctrl = Profile.controller()

    def tearDown(self):
        self._ctrl.unselect()
        self.resetStorage()

    # convenience helper -----------------------------------------------
    def _make(self, name, **kwargs):
        return self._ctrl.create({'name': name, **kwargs})

    # ------------------------------------------------------------------

    def test_create_and_one_by_name(self):
        self._make('p1')
        result = self._ctrl.one({'name': 'p1'})
        self.assertIsNotNone(result)
        self.assertEqual(result['name'], 'p1')

    def test_one_missing_returns_none(self):
        self.assertIsNone(self._ctrl.one({'name': 'noexist'}))

    def test_all_empty(self):
        self.assertEqual(self._ctrl.all(), [])

    def test_all_returns_all_created(self):
        self._make('a')
        self._make('b')
        names = {p['name'] for p in self._ctrl.all()}
        self.assertEqual(names, {'a', 'b'})

    def test_count_empty(self):
        self.assertEqual(self._ctrl.count(), 0)

    def test_count_reflects_inserts(self):
        self._make('c1')
        self.assertEqual(self._ctrl.count(), 1)
        self._make('c2')
        self.assertEqual(self._ctrl.count(), 2)

    def test_exists_true(self):
        self._make('ex')
        self.assertTrue(self._ctrl.exists({'name': 'ex'}))

    def test_exists_false(self):
        self.assertFalse(self._ctrl.exists({'name': 'nx'}))

    def test_delete_removes_record(self):
        self._make('del_me')
        self._ctrl.delete({'name': 'del_me'})
        self.assertIsNone(self._ctrl.one({'name': 'del_me'}))

    def test_unique_constraint_on_create(self):
        """Creating two records with the same primary key raises UniqueAttributeError."""
        self._make('unique')
        with self.assertRaises(UniqueAttributeError):
            self._make('unique')


class ControllerSearchMatchTest(tests.TestCase):
    """search / match."""

    def setUp(self):
        self.resetStorage()
        self._ctrl = Profile.controller()
        self._ctrl.create({'name': 'docker_profile', 'backend': 'docker'})
        self._ctrl.create({'name': 'podman_profile', 'backend': 'podman'})
        self._ctrl.create({'name': 'apptainer_profile', 'backend': 'apptainer'})

    def tearDown(self):
        self._ctrl.unselect()
        self.resetStorage()

    def test_search_by_backend(self):
        results = self._ctrl.search({'backend': 'docker'})
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0]['name'], 'docker_profile')

    def test_search_no_match_returns_empty(self):
        results = self._ctrl.search({'backend': 'shifter'})
        self.assertEqual(results, [])

    def test_match_by_name_prefix_regex(self):
        results = self._ctrl.match('name', regex='^docker')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0]['name'], 'docker_profile')

    def test_match_by_name_ends_with(self):
        # TinyDB uses re.match() which anchors at the start, so use a test
        # callable to check suffix membership instead of a bare suffix regex.
        results = self._ctrl.match('name',
                                   test=lambda v: v.endswith('_profile'))
        self.assertEqual(len(results), 3)

    def test_match_by_test_function(self):
        results = self._ctrl.match('backend',
                                   test=lambda v: v in {'docker', 'podman'})
        names = {r['name'] for r in results}
        self.assertEqual(names, {'docker_profile', 'podman_profile'})


class ControllerUpdateTest(tests.TestCase):
    """update / unset."""

    def setUp(self):
        self.resetStorage()
        self._ctrl = Profile.controller()

    def tearDown(self):
        self._ctrl.unselect()
        self.resetStorage()

    def test_update_changes_field(self):
        self._ctrl.create({'name': 'upd', 'backend': 'singularity'})
        self._ctrl.update({'backend': 'apptainer'}, {'name': 'upd'})
        result = self._ctrl.one({'name': 'upd'})
        self.assertEqual(result['backend'], 'apptainer')

    def test_update_nonexistent_raises_model_error(self):
        with self.assertRaises(ModelError):
            self._ctrl.update({'backend': 'x'}, {'name': 'ghost'})

    def test_update_invalid_attribute_raises_model_error(self):
        self._ctrl.create({'name': 'p'})
        with self.assertRaises(ModelError):
            self._ctrl.update({'nonexistent_attr': 'val'}, {'name': 'p'})

    def test_unset_removes_field(self):
        self._ctrl.create({'name': 'p', 'backend': 'docker', 'image': 'foo.sif'})
        self._ctrl.unset(['image'], {'name': 'p'})
        result = self._ctrl.one({'name': 'p'})
        self.assertNotIn('image', result)

    def test_unset_invalid_attr_raises_model_error(self):
        self._ctrl.create({'name': 'q'})
        with self.assertRaises(ModelError):
            self._ctrl.unset(['completely_bogus_field'], {'name': 'q'})


class ControllerTopicBusTest(tests.TestCase):
    """push_to_topic / pop_topic message bus."""

    def test_push_and_pop(self):
        Profile.controller().push_to_topic('t', 'msg1')
        Profile.controller().push_to_topic('t', 'msg2')
        msgs = Profile.controller().pop_topic('t')
        self.assertIn('msg1', msgs)
        self.assertIn('msg2', msgs)

    def test_second_pop_returns_empty(self):
        Profile.controller().push_to_topic('t2', 'x')
        Profile.controller().pop_topic('t2')
        self.assertEqual(Profile.controller().pop_topic('t2'), [])

    def test_pop_nonexistent_topic_returns_empty(self):
        self.assertEqual(Profile.controller().pop_topic('no_such_topic'), [])



"""
Version helper module
"""

import re
from e4s_cl.error import InternalError


class Version(list):
    """
    Class abstracting a version of form x.y.z.
    Comparison operators are defined for simplicity.
    """

    def __init__(self, string):
        super().__init__()

        digits = re.search(r'(?P<version>[0-9]+(\.[0-9]+)+)', string)

        if digits:
            for digit in digits.group('version').split('.'):
                self.append(int(digit))

    def __str__(self):
        return ".".join([str(digit) for digit in self])

    def __bool__(self):
        return bool(len(self))

    def __gt__(self, rhs):
        if not isinstance(rhs, Version):
            raise InternalError(
                f"Invalid operation > with object of type {type(rhs)}")

        for lhs_d, rhs_d in zip(self, rhs):
            if lhs_d != rhs_d:
                return lhs_d > rhs_d

        return False

    @property
    def major(self):
        return self[0]

    @property
    def minor(self):
        return self[1]

    @property
    def patch(self):
        return self[2]



"""
Test launcher module
"""

from e4s_cl.cf.launchers import Parser

SCRIPT_NAMES = ['mylauncher']

ARGUMENTS = {
    "-a": 0,
    "-b": 1,
    "-c": 5,
}

META = dict(reserved_directories=['/reserved'])

PARSER = Parser(ARGUMENTS)
